r"""
Internal functions not intended for use by users. This may change without
notice.
"""
from .. import logger

from ..modules.vtkPVInSitu import vtkInSituPipelinePython

from ..modules.vtkPVPythonCatalyst import vtkCPPythonScriptV2Helper


def _get_active_helper():
    return vtkCPPythonScriptV2Helper.GetActiveInstance()


def _get_active_data_description():
    if not IsInsitu():
        return None
    helper = _get_active_helper()
    return helper.GetDataDescription()


def _transform_registration_name(name):
    from . import get_args
    import re, argparse

    regex = re.compile(r"^\${args\.([\w.-]+)}$")
    m = regex.match(name)
    if not m:
        return name

    argname = m.group(1)
    parser = argparse.ArgumentParser()
    parser.add_argument("--%s" % argname, dest=argname)
    parser.add_argument("-%s" % argname, dest=argname)
    result = parser.parse_known_args(get_args())[0]
    val = getattr(result, argname)
    return val if val else name


def IsInsitu():
    """Returns True if executing in an insitu environment, else false"""
    helper = _get_active_helper()
    return (helper is not None)


def IsLegacyCatalystAdaptor():
    """Returns True if the active execution environment is from within a legacy
    Catalyst adaptor implementation that uses vtkCPProcessor etc. manually,
    instead of the Conduit-based in situ API"""
    return _get_active_data_description() is not None


if IsInsitu() and not IsLegacyCatalystAdaptor():
    from ..modules.vtkPVInSitu import vtkInSituInitializationHelper


def IsCatalystInSituAPI():
    """Returns True if the active execution environment is from within
    an implementation of the Conduit-based Catalyst In Situ API."""
    return IsInsitu() and not IsLegacyCatalystAdaptor() and vtkInSituInitializationHelper.IsInitialized()


def IsInsituInput(name):
    if not name or not IsInsitu():
        return False
    name = _transform_registration_name(name)
    dataDesc = _get_active_data_description()
    if IsLegacyCatalystAdaptor():
        # Legacy Catalyst
        if dataDesc.GetInputDescriptionByName(name) is not None:
            return True
    elif IsCatalystInSituAPI() and (vtkInSituInitializationHelper.GetProducer(name) is not None):
        # Catalyst 2.0
        return True
    return False


def RegisterExtractor(extractor):
    """Keeps track of extractors created inside a specific Catalyst
    script.  This is useful to ensure we only update the extractors for that
    current script when multiple scripts are being executed in the same run.
    """
    assert IsInsitu()
    _get_active_helper().RegisterExtractor(extractor.SMProxy)


def RegisterView(view):
    """Keeps track of views created inside a specific Catalyst
    script.  This is useful to ensure we only update the views for the
    current script when multiple scripts are being executed in the same run.
    """
    assert IsInsitu()
    _get_active_helper().RegisterView(view.SMProxy)

    if IsCatalystInSituAPI():
        view.ViewTime = vtkInSituInitializationHelper.GetTime()
    else:
        view.ViewTime = _get_active_data_description().GetTime()


def CreateProducer(name):
    assert IsInsituInput(name)
    from . import log_level
    from .. import log

    originalname = name
    name = _transform_registration_name(name)
    log(log_level(), "creating producer for simulation input named '%s' (original-name=%s)" \
        % (name, originalname))
    if IsCatalystInSituAPI():
        # Catalyst 2.0
        from paraview import servermanager
        producer = servermanager._getPyProxy(vtkInSituInitializationHelper.GetProducer(name))

        # since state file may have arbitrary properties being specified
        # on the original source, we ensure we ignore them
        producer.IgnoreUnknownSetRequests = True
        return producer

    # Legacy Catalyst
    from paraview import servermanager
    helper = _get_active_helper()
    producer = helper.GetTrivialProducer(name)
    producer = servermanager._getPyProxy(producer)

    # since state file may have arbitrary properties being specified
    # on the original source, we ensure we ignore them
    producer.IgnoreUnknownSetRequests = True
    return producer


def InitializePythonEnvironment():
    """
    This is called in vtkCPPythonPipeline to initialize the Python
    environment to be more suitable for Catalyst in situ execution.
    """
    from . import log_level
    from .. import log

    log(log_level(), "initializing Python environment for in situ")

    # disable writing bytecode, we don't want to generate pyc files
    # when running in situ.
    log(log_level(), "disable byte-code generation")
    import sys
    sys.dont_write_bytecode = True

    # ensure we can import ParaView servermanager and necessary modules
    log(log_level(), "import paraview.servermanager")
    from paraview import servermanager

    # import Python wrapping for vtkPVCatalyst module
    log(log_level(), "import paraview.modules.vtkPVCatalyst")
    from paraview.modules import vtkPVCatalyst


def RegisterPackageFromZip(zipfilename, packagename=None):
    from . import importers
    zipfilename = _mpi_exchange_if_needed(zipfilename)
    return importers.add_file(zipfilename, packagename)


def RegisterPackageFromDir(path):
    import os.path
    from . import importers
    packagename = os.path.basename(path)
    init_py = os.path.join(path, "__init__.py")
    return importers.add_file(init_py, packagename)


def RegisterModuleFromFile(filename):
    from . import importers
    return importers.add_file(filename)


_temp_directory = None


def _mpi_exchange_if_needed(filename):
    global _temp_directory

    from ..modules.vtkRemotingCore import vtkProcessModule
    pm = vtkProcessModule.GetProcessModule()
    if pm.GetNumberOfLocalPartitions() <= 1:
        return filename

    from mpi4py import MPI
    from vtkmodules.vtkParallelMPI4Py import vtkMPI4PyCommunicator
    comm = vtkMPI4PyCommunicator.ConvertToPython(pm.GetGlobalController().GetCommunicator())
    if comm.Get_rank() == 0:
        with open(filename, 'rb') as f:
            data = f.read()
    else:
        data = None
    data = comm.bcast(data, root=0)
    if comm.Get_rank() == 0:
        return filename
    else:
        if not _temp_directory:
            import tempfile, os.path
            # we hook the temp-dir to the module so it lasts for the livetime of
            # the interpreter and gets cleaned up on exit.
            _temp_directory = tempfile.TemporaryDirectory()
        filename = os.path.join(_temp_directory.name, os.path.basename(filename))
        with open(filename, "wb") as f:
            f.write(data)
        return filename



from vtkmodules.vtkIOImage import vtkPNGReader
from vtkmodules.vtkCommonCore import vtkFloatArray, vtkUnsignedCharArray
from vtkmodules.vtkCommonDataModel import vtkImageData
from vtkmodules.vtkIOLegacy import vtkDataSetWriter

from vtkmodules.web.camera import normalize, vectProduct, dotProduct
from vtkmodules.web import iteritems

import json, os, math, array

# -----------------------------------------------------------------------------
# Helper function
# -----------------------------------------------------------------------------


def getScalarFromRGB(rgb, scalarRange=[-1.0, 1.0]):
    delta = (scalarRange[1] - scalarRange[0]) / 16777215.0  # 2^24 - 1 => 16,777,215
    if rgb[0] != 0 or rgb[1] != 0 or rgb[2] != 0:
        # Decode encoded value
        return scalarRange[0] + delta * float(
            rgb[0] * 65536 + rgb[1] * 256 + rgb[2] - 1
        )
    else:
        # No value
        return float("NaN")


def convertImageToFloat(srcPngImage, destFile, scalarRange=[0.0, 1.0]):
    reader = vtkPNGReader()
    reader.SetFileName(srcPngImage)
    reader.Update()
    rgbArray = reader.GetOutput().GetPointData().GetArray(0)
    stackSize = rgbArray.GetNumberOfTuples()
    size = reader.GetOutput().GetDimensions()

    outputArray = vtkFloatArray()
    outputArray.SetNumberOfComponents(1)
    outputArray.SetNumberOfTuples(stackSize)

    for idx in range(stackSize):
        outputArray.SetTuple1(
            idx, getScalarFromRGB(rgbArray.GetTuple(idx), scalarRange)
        )

    # Write float file
    with open(destFile, "wb") as f:
        f.write(memoryview(outputArray))

    return size


def convertRGBArrayToFloatArray(rgbArray, scalarRange=[0.0, 1.0]):
    linearSize = rgbArray.GetNumberOfTuples()

    outputArray = vtkFloatArray()
    outputArray.SetNumberOfComponents(1)
    outputArray.SetNumberOfTuples(linearSize)

    for idx in range(linearSize):
        outputArray.SetTuple1(
            idx, getScalarFromRGB(rgbArray.GetTuple(idx), scalarRange)
        )

    return outputArray


# -----------------------------------------------------------------------------
# Composite.json To order.array
# -----------------------------------------------------------------------------
class CompositeJSON(object):
    def __init__(self, numberOfLayers):
        self.nbLayers = numberOfLayers
        self.encoding = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    def load(self, file):
        with open(file, "r") as f:
            composite = json.load(f)
            self.width = composite["dimensions"][0]
            self.height = composite["dimensions"][1]
            self.pixels = composite["pixel-order"].split("+")
            self.imageSize = self.width * self.height
            self.stackSize = self.imageSize * self.nbLayers

    def getImageSize(self):
        return self.imageSize

    def getStackSize(self):
        return self.stackSize

    def writeOrderSprite(self, path):
        ds = vtkImageData()
        ds.SetDimensions(self.width, self.height, self.nbLayers)
        ds.GetPointData().AddArray(self.getSortedOrderArray())

        writer = vtkDataSetWriter()
        writer.SetInputData(ds)
        writer.SetFileName(path)
        writer.Update()

    def getSortedOrderArray(self):
        sortedOrder = vtkUnsignedCharArray()
        sortedOrder.SetName("layerIdx")
        sortedOrder.SetNumberOfTuples(self.stackSize)

        # Reset content
        for idx in range(self.stackSize):
            sortedOrder.SetValue(idx, 255)

        idx = 0
        for pixel in self.pixels:
            x = idx % self.width
            y = idx / self.width
            flipYIdx = self.width * (self.height - y - 1) + x
            if "@" in pixel:
                idx += int(pixel[1:])
            else:
                # Need to decode the order
                layerIdx = 0
                for layer in pixel:
                    sortedOrder.SetValue(
                        flipYIdx + self.imageSize * layerIdx, self.encoding.index(layer)
                    )
                    layerIdx += 1

                # Move to next pixel
                idx += 1

        return sortedOrder


# -----------------------------------------------------------------------------
# Composite Sprite to Sorted Composite Dataset Builder
# -----------------------------------------------------------------------------


class ConvertCompositeSpriteToSortedStack(object):
    def __init__(self, directory):
        self.basePath = directory
        self.layers = []
        self.data = []
        self.imageReader = vtkPNGReader()

        # Load JSON metadata
        with open(os.path.join(directory, "config.json"), "r") as f:
            self.config = json.load(f)
            self.nbLayers = len(self.config["scene"])
            while len(self.layers) < self.nbLayers:
                self.layers.append({})

        with open(os.path.join(directory, "index.json"), "r") as f:
            self.info = json.load(f)

        with open(os.path.join(directory, "offset.json"), "r") as f:
            offsets = json.load(f)
            for key, value in iteritems(offsets):
                meta = key.split("|")
                if len(meta) == 2:
                    self.layers[int(meta[0])][meta[1]] = value
                elif meta[1] in self.layers[int(meta[0])]:
                    self.layers[int(meta[0])][meta[1]][int(meta[2])] = value
                else:
                    self.layers[int(meta[0])][meta[1]] = [value, value, value]

        self.composite = CompositeJSON(len(self.layers))

    def listData(self):
        return self.data

    def convert(self):
        for root, dirs, files in os.walk(self.basePath):
            if "rgb.png" in files:
                print("Process", root)
                self.processDirectory(root)

    def processDirectory(self, directory):
        self.imageReader.SetFileName(os.path.join(directory, "rgb.png"))
        self.imageReader.Update()
        rgbArray = self.imageReader.GetOutput().GetPointData().GetArray(0)

        self.composite.load(os.path.join(directory, "composite.json"))
        orderArray = self.composite.getSortedOrderArray()

        imageSize = self.composite.getImageSize()
        stackSize = self.composite.getStackSize()

        # Write order (sorted order way)
        with open(os.path.join(directory, "order.uint8"), "wb") as f:
            f.write(memoryview(orderArray))
            self.data.append(
                {"name": "order", "type": "array", "fileName": "/order.uint8"}
            )

        # Encode Normals (sorted order way)
        if "normal" in self.layers[0]:
            sortedNormal = vtkUnsignedCharArray()
            sortedNormal.SetNumberOfComponents(3)  # x,y,z
            sortedNormal.SetNumberOfTuples(stackSize)

            # Get Camera orientation and rotation information
            camDir = [0, 0, 0]
            worldUp = [0, 0, 0]
            with open(os.path.join(directory, "camera.json"), "r") as f:
                camera = json.load(f)
                camDir = normalize(
                    [camera["position"][i] - camera["focalPoint"][i] for i in range(3)]
                )
                worldUp = normalize(camera["viewUp"])

            # [ camRight, camUp, camDir ] will be our new orthonormal basis for normals
            camRight = vectProduct(camDir, worldUp)
            camUp = vectProduct(camRight, camDir)

            # Tmp structure to capture (x,y,z) normal
            normalByLayer = vtkFloatArray()
            normalByLayer.SetNumberOfComponents(3)
            normalByLayer.SetNumberOfTuples(stackSize)

            # Capture all layer normals
            layerIdx = 0
            zPosCount = 0
            zNegCount = 0
            for layer in self.layers:
                normalOffset = layer["normal"]
                for idx in range(imageSize):
                    normalByLayer.SetTuple3(
                        layerIdx * imageSize + idx,
                        getScalarFromRGB(
                            rgbArray.GetTuple(idx + normalOffset[0] * imageSize)
                        ),
                        getScalarFromRGB(
                            rgbArray.GetTuple(idx + normalOffset[1] * imageSize)
                        ),
                        getScalarFromRGB(
                            rgbArray.GetTuple(idx + normalOffset[2] * imageSize)
                        ),
                    )

                    # Re-orient normal to be view based
                    vect = normalByLayer.GetTuple3(layerIdx * imageSize + idx)
                    if not math.isnan(vect[0]):
                        # Express normal in new basis we computed above
                        rVect = normalize(
                            [
                                -dotProduct(vect, camRight),
                                dotProduct(vect, camUp),
                                dotProduct(vect, camDir),
                            ]
                        )

                        # Need to reverse vector ?
                        if rVect[2] < 0:
                            normalByLayer.SetTuple3(
                                layerIdx * imageSize + idx,
                                -rVect[0],
                                -rVect[1],
                                -rVect[2],
                            )
                        else:
                            normalByLayer.SetTuple3(
                                layerIdx * imageSize + idx, rVect[0], rVect[1], rVect[2]
                            )

                layerIdx += 1

            # Sort normals and encode them as 3 bytes ( -1 < xy < 1 | 0 < z < 1)
            for idx in range(stackSize):
                layerIdx = int(orderArray.GetValue(idx))
                if layerIdx == 255:
                    # No normal => same as view direction
                    sortedNormal.SetTuple3(idx, 128, 128, 255)
                else:
                    offset = layerIdx * imageSize
                    imageIdx = idx % imageSize
                    vect = normalByLayer.GetTuple3(imageIdx + offset)
                    if (
                        not math.isnan(vect[0])
                        and not math.isnan(vect[1])
                        and not math.isnan(vect[2])
                    ):
                        sortedNormal.SetTuple3(
                            idx,
                            int(127.5 * (vect[0] + 1)),
                            int(127.5 * (vect[1] + 1)),
                            int(255 * vect[2]),
                        )
                    else:
                        print(
                            "WARNING: encountered NaN in normal of layer ",
                            layerIdx,
                            ": [",
                            vect[0],
                            ",",
                            vect[1],
                            ",",
                            vect[2],
                            "]",
                        )
                        sortedNormal.SetTuple3(idx, 128, 128, 255)

            # Write the sorted data
            with open(os.path.join(directory, "normal.uint8"), "wb") as f:
                f.write(memoryview(sortedNormal))
                self.data.append(
                    {
                        "name": "normal",
                        "type": "array",
                        "fileName": "/normal.uint8",
                        "categories": ["normal"],
                    }
                )

        # Encode Intensity (sorted order way)
        if "intensity" in self.layers[0]:
            intensityOffsets = []
            sortedIntensity = vtkUnsignedCharArray()
            sortedIntensity.SetNumberOfTuples(stackSize)

            for layer in self.layers:
                intensityOffsets.append(layer["intensity"])

            for idx in range(stackSize):
                layerIdx = int(orderArray.GetValue(idx))
                if layerIdx == 255:
                    sortedIntensity.SetValue(idx, 255)
                else:
                    offset = 3 * intensityOffsets[layerIdx] * imageSize
                    imageIdx = idx % imageSize
                    sortedIntensity.SetValue(
                        idx, rgbArray.GetValue(imageIdx * 3 + offset)
                    )

            with open(os.path.join(directory, "intensity.uint8"), "wb") as f:
                f.write(memoryview(sortedIntensity))
                self.data.append(
                    {
                        "name": "intensity",
                        "type": "array",
                        "fileName": "/intensity.uint8",
                        "categories": ["intensity"],
                    }
                )

        # Encode Each layer Scalar
        layerIdx = 0
        for layer in self.layers:
            for scalar in layer:
                if scalar not in ["intensity", "normal"]:
                    offset = imageSize * layer[scalar]
                    scalarRange = self.config["scene"][layerIdx]["colors"][scalar][
                        "range"
                    ]
                    delta = (
                        scalarRange[1] - scalarRange[0]
                    ) / 16777215.0  # 2^24 - 1 => 16,777,215

                    scalarArray = vtkFloatArray()
                    scalarArray.SetNumberOfTuples(imageSize)
                    for idx in range(imageSize):
                        rgb = rgbArray.GetTuple(idx + offset)
                        if rgb[0] != 0 or rgb[1] != 0 or rgb[2] != 0:
                            # Decode encoded value
                            value = scalarRange[0] + delta * float(
                                rgb[0] * 65536 + rgb[1] * 256 + rgb[2] - 1
                            )
                            scalarArray.SetValue(idx, value)
                        else:
                            # No value
                            scalarArray.SetValue(idx, float("NaN"))

                    with open(
                        os.path.join(directory, "%d_%s.float32" % (layerIdx, scalar)),
                        "wb",
                    ) as f:
                        f.write(memoryview(scalarArray))
                        self.data.append(
                            {
                                "name": "%d_%s" % (layerIdx, scalar),
                                "type": "array",
                                "fileName": "/%d_%s.float32" % (layerIdx, scalar),
                                "categories": ["%d_%s" % (layerIdx, scalar)],
                            }
                        )

            layerIdx += 1


# -----------------------------------------------------------------------------
# Composite Sprite to Sorted Composite Dataset Builder
# -----------------------------------------------------------------------------


class ConvertCompositeDataToSortedStack(object):
    def __init__(self, directory):
        self.basePath = directory
        self.layers = []
        self.data = []
        self.imageReader = vtkPNGReader()

        # Load JSON metadata
        with open(os.path.join(directory, "config.json"), "r") as f:
            self.config = json.load(f)
            self.nbLayers = len(self.config["scene"])
            while len(self.layers) < self.nbLayers:
                self.layers.append({})

        with open(os.path.join(directory, "index.json"), "r") as f:
            self.info = json.load(f)

    def listData(self):
        return self.data

    def convert(self):
        for root, dirs, files in os.walk(self.basePath):
            if "depth_0.float32" in files:
                print("Process", root)
                self.processDirectory(root)

    def processDirectory(self, directory):
        # Load depth
        depthStack = []
        imageSize = self.config["size"]
        linearSize = imageSize[0] * imageSize[1]
        nbLayers = len(self.layers)
        stackSize = nbLayers * linearSize
        layerList = range(nbLayers)
        for layerIdx in layerList:
            with open(
                os.path.join(directory, "depth_%d.float32" % layerIdx), "rb"
            ) as f:
                a = array.array("f")
                a.fromfile(f, linearSize)
                depthStack.append(a)

        orderArray = vtkUnsignedCharArray()
        orderArray.SetName("layerIdx")
        orderArray.SetNumberOfComponents(1)
        orderArray.SetNumberOfTuples(stackSize)

        pixelSorter = [(i, i) for i in layerList]

        for pixelId in range(linearSize):
            # Fill pixelSorter
            for layerIdx in layerList:
                if depthStack[layerIdx][pixelId] < 1.0:
                    pixelSorter[layerIdx] = (layerIdx, depthStack[layerIdx][pixelId])
                else:
                    pixelSorter[layerIdx] = (255, 1.0)

            # Sort pixel layers
            pixelSorter.sort(key=lambda tup: tup[1])

            # Fill sortedOrder array
            for layerIdx in layerList:
                orderArray.SetValue(
                    layerIdx * linearSize + pixelId, pixelSorter[layerIdx][0]
                )

        # Write order (sorted order way)
        with open(os.path.join(directory, "order.uint8"), "wb") as f:
            f.write(memoryview(orderArray))
            self.data.append(
                {"name": "order", "type": "array", "fileName": "/order.uint8"}
            )

        # Remove depth files
        for layerIdx in layerList:
            os.remove(os.path.join(directory, "depth_%d.float32" % layerIdx))

        # Encode Normals (sorted order way)
        if "normal" in self.config["light"]:
            sortedNormal = vtkUnsignedCharArray()
            sortedNormal.SetNumberOfComponents(3)  # x,y,z
            sortedNormal.SetNumberOfTuples(stackSize)

            # Get Camera orientation and rotation information
            camDir = [0, 0, 0]
            worldUp = [0, 0, 0]
            with open(os.path.join(directory, "camera.json"), "r") as f:
                camera = json.load(f)
                camDir = normalize(
                    [camera["position"][i] - camera["focalPoint"][i] for i in range(3)]
                )
                worldUp = normalize(camera["viewUp"])

            # [ camRight, camUp, camDir ] will be our new orthonormal basis for normals
            camRight = vectProduct(camDir, worldUp)
            camUp = vectProduct(camRight, camDir)

            # Tmp structure to capture (x,y,z) normal
            normalByLayer = vtkFloatArray()
            normalByLayer.SetNumberOfComponents(3)
            normalByLayer.SetNumberOfTuples(stackSize)

            # Capture all layer normals
            zPosCount = 0
            zNegCount = 0
            for layerIdx in layerList:

                # Load normal(x,y,z) from current layer
                normalLayer = []
                for comp in [0, 1, 2]:
                    with open(
                        os.path.join(
                            directory, "normal_%d_%d.float32" % (layerIdx, comp)
                        ),
                        "rb",
                    ) as f:
                        a = array.array("f")
                        a.fromfile(f, linearSize)
                        normalLayer.append(a)

                # Store normal inside vtkArray
                offset = layerIdx * linearSize
                for idx in range(linearSize):
                    normalByLayer.SetTuple3(
                        idx + offset,
                        normalLayer[0][idx],
                        normalLayer[1][idx],
                        normalLayer[2][idx],
                    )

                    # Re-orient normal to be view based
                    vect = normalByLayer.GetTuple3(layerIdx * linearSize + idx)
                    if not math.isnan(vect[0]):
                        # Express normal in new basis we computed above
                        rVect = normalize(
                            [
                                -dotProduct(vect, camRight),
                                dotProduct(vect, camUp),
                                dotProduct(vect, camDir),
                            ]
                        )

                        # Need to reverse vector ?
                        if rVect[2] < 0:
                            normalByLayer.SetTuple3(
                                layerIdx * linearSize + idx,
                                -rVect[0],
                                -rVect[1],
                                -rVect[2],
                            )
                        else:
                            normalByLayer.SetTuple3(
                                layerIdx * linearSize + idx,
                                rVect[0],
                                rVect[1],
                                rVect[2],
                            )

            # Sort normals and encode them as 3 bytes ( -1 < xy < 1 | 0 < z < 1)
            for idx in range(stackSize):
                layerIdx = int(orderArray.GetValue(idx))
                if layerIdx == 255:
                    # No normal => same as view direction
                    sortedNormal.SetTuple3(idx, 128, 128, 255)
                else:
                    offset = layerIdx * linearSize
                    imageIdx = idx % linearSize
                    vect = normalByLayer.GetTuple3(imageIdx + offset)
                    if (
                        not math.isnan(vect[0])
                        and not math.isnan(vect[1])
                        and not math.isnan(vect[2])
                    ):
                        sortedNormal.SetTuple3(
                            idx,
                            int(127.5 * (vect[0] + 1)),
                            int(127.5 * (vect[1] + 1)),
                            int(255 * vect[2]),
                        )
                    else:
                        print(
                            "WARNING: encountered NaN in normal of layer ",
                            layerIdx,
                            ": [",
                            vect[0],
                            ",",
                            vect[1],
                            ",",
                            vect[2],
                            "]",
                        )
                        sortedNormal.SetTuple3(idx, 128, 128, 255)

            # Write the sorted data
            with open(os.path.join(directory, "normal.uint8"), "wb") as f:
                f.write(memoryview(sortedNormal))
                self.data.append(
                    {
                        "name": "normal",
                        "type": "array",
                        "fileName": "/normal.uint8",
                        "categories": ["normal"],
                    }
                )

            # Remove depth files
            for layerIdx in layerList:
                os.remove(
                    os.path.join(directory, "normal_%d_%d.float32" % (layerIdx, 0))
                )
                os.remove(
                    os.path.join(directory, "normal_%d_%d.float32" % (layerIdx, 1))
                )
                os.remove(
                    os.path.join(directory, "normal_%d_%d.float32" % (layerIdx, 2))
                )

        # Encode Intensity (sorted order way)
        if "intensity" in self.config["light"]:
            sortedIntensity = vtkUnsignedCharArray()
            sortedIntensity.SetNumberOfTuples(stackSize)

            intensityLayers = []
            for layerIdx in layerList:
                with open(
                    os.path.join(directory, "intensity_%d.uint8" % layerIdx), "rb"
                ) as f:
                    a = array.array("B")
                    a.fromfile(f, linearSize)
                    intensityLayers.append(a)

            for idx in range(stackSize):
                layerIdx = int(orderArray.GetValue(idx))
                if layerIdx == 255:
                    sortedIntensity.SetValue(idx, 255)
                else:
                    imageIdx = idx % linearSize
                    sortedIntensity.SetValue(idx, intensityLayers[layerIdx][imageIdx])

            with open(os.path.join(directory, "intensity.uint8"), "wb") as f:
                f.write(memoryview(sortedIntensity))
                self.data.append(
                    {
                        "name": "intensity",
                        "type": "array",
                        "fileName": "/intensity.uint8",
                        "categories": ["intensity"],
                    }
                )

            # Remove depth files
            for layerIdx in layerList:
                os.remove(os.path.join(directory, "intensity_%d.uint8" % layerIdx))



# SPDX-FileCopyrightText: Copyright (c) Kitware Inc.
# SPDX-License-Identifier: BSD-3-Clause

r"""python_view is a module providing access to a PythonView. It is
possible to use the PythonView API directly, but this module provides
convenience classes in Python.
"""
import paraview
from vtkmodules.vtkCommonDataModel import vtkImageData


def numpy_to_image(numpy_array):
    """Convert a numpy 2D or 3D array to a vtkImageData object.

    numpy_array
      2D or 3D numpy array containing image data

    return
      vtkImageData with the numpy_array content
    """
    try:
        import numpy
    except:
        paraview.print_error("Error: Cannot import numpy")

    shape = numpy_array.shape
    if len(shape) < 2:
        raise Exception('numpy array must have dimensionality of at least 2')

    h, w = shape[0], shape[1]
    c = 1
    if len(shape) == 3:
        c = shape[2]

    # Reshape 2D image to 1D array suitable for conversion to a
    # vtkArray with numpy_support.numpy_to_vtk()
    linear_array = numpy.reshape(numpy_array, (w * h, c))

    try:
        from vtkmodules.util import numpy_support
    except:
        paraview.print_error("Error: Cannot import vtkmodules.util.numpy_support")

    vtk_array = numpy_support.numpy_to_vtk(linear_array)

    image = vtkImageData()
    image.SetDimensions(w, h, 1)
    image.AllocateScalars(vtk_array.GetDataType(), 4)
    image.GetPointData().GetScalars().DeepCopy(vtk_array)

    return image


def figure_to_data(figure):
    """Convert a Matplotlib figure to a numpy 2D array with RGBA uint8 channels
    and return it.

    figure
      A matplotlib figure.

    return
      A numpy 2D array of RGBA values.
    """
    # Draw the renderer
    try:
        import matplotlib
    except:
        paraview.print_error("Error: Cannot import matplotlib")

    figure.canvas.draw()

    # Get the RGBA buffer from the figure
    w, h = figure.canvas.get_width_height()

    try:
        import numpy
    except:
        paraview.print_error("Error: Cannot import numpy")

    buf = numpy.fromstring(figure.canvas.tostring_argb(), dtype=numpy.uint8)
    buf.shape = (h, w, 4)

    # canvas.tostring_argb gives pixmap in ARGB mode. Roll the alpha channel to have it in RGBA mode
    buf = numpy.roll(buf, 3, axis=2)

    return buf


def figure_to_image(figure):
    """Convert a Matplotlib figure to a vtkImageData with RGBA unsigned char
    channels.

    figure
      A matplotlib figure.

    return
      A vtkImageData with the Matplotlib figure content.
    """
    buf = figure_to_data(figure)

    # Flip rows to be suitable for vtkImageData.
    buf = buf[::-1, :, :].copy()

    return numpy_to_image(buf)


def matplotlib_figure(width, height):
    """Create a Matplotlib figure with specified width and height for rendering.

    w
      Width of desired plot.
    h
      Height of desired plot.

    return
      A Matplotlib figure.
    """
    try:
        from matplotlib.backends.backend_agg import FigureCanvasAgg
    except:
        paraview.print_error("Error: Cannot import matplotlib.backends.backend_agg.FigureCanvasAgg")
    try:
        from matplotlib.figure import Figure
    except:
        paraview.print_error("Error: Cannot import matplotlib.figure.Figure")

    figure = Figure()
    figureCanvas = FigureCanvasAgg(figure)
    figure.set_dpi(72)
    figure.set_size_inches(float(width) / 72.0, float(height) / 72.0)

    return figure


def call_render(render_function, view, width, height):
    """Utility function to call the user-defined render function. This is called
    by the C++ side of the vtkPythonView class.

    view
      vtkPythonView object

    width
      Width of view

    height
      Height of view
    """
    if render_function == None:
        return

    # Check how many parameters it takes.
    num_args = render_function.__code__.co_argcount

    image = None
    if (num_args == 3):
        # Current-style render() function
        image = render_function(view, width, height)

    elif (num_args == 2):
        # Old-style render() function introduced in ParaView 4.1
        figure = matplotlib_figure(width, height)
        render_function(view, figure)

        image = figure_to_image(figure)

    view.SetImageData(image)



import argparse, json, os, os.path, sys


def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def find_subdirectory(subdir, node):
    contents = node["contents"]
    for item in contents:
        if item["type"] == "directory" and item["name"] == subdir:
            return item
    raise RuntimeError("Missing subdir '%s'" % subdir)


def find_directory(subdir, node):
    if subdir == '.' or subdir == None:
        return node
    for part in splitall(subdir):
        node = find_subdirectory(part, node)
    return node


def get_baseline(tname, path, image, root):
    import re
    filename = os.path.join(path, image)
    relpath = os.path.relpath(filename, root)
    return tname + "_" + re.sub('[^0-9a-zA-Z.]+', '_', relpath)


def _get_options():
    parser = argparse.ArgumentParser( \
        description="Validate Extrats generated in tests")

    parser.add_argument("--name", type=str, required=True,
                        help="name of the test being validated")
    parser.add_argument("--subdir", type=str, default=None,
                        help="sub directory to validate")
    parser.add_argument("--root", type=str, required=True,
                        help="root directory with results")
    parser.add_argument("--json", type=str, required=True,
                        help="validation json file")
    parser.add_argument("--baseline-dir", type=str, default=None,
                        help="directory for baselines for regression tests, if any")
    parser.add_argument("--temp-dir", type=str, default="/tmp",
                        help="directory for results for regression test failures, if any")

    return parser


def main():
    parser = _get_options()
    args = parser.parse_args()

    with open(args.json, 'r') as f:
        data = json.load(f)[0]

    node = data
    test_dir = args.root
    if args.subdir is not None:
        node = find_directory(args.subdir, node)
        test_dir = os.path.join(args.root, args.subdir)

    # Check that directory exists since `os.walk()` does not fail
    # for non extant directories.
    if not os.path.isdir(test_dir):
        raise RuntimeError("Missing directory: '%s'" % test_dir)

    # Validate directory structure.
    regressions_failure_count = 0
    for dirpath, dirs, files in os.walk(test_dir):
        # skip hidden files and directories
        # ref: https://stackoverflow.com/questions/13454164/os-walk-without-hidden-folders
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        relpath = os.path.relpath(dirpath, test_dir)
        currentnode = find_directory(relpath, node)
        if not currentnode:
            raise RuntimeError("Unexpected directory was found: '%s'" % dirpath)

        dirs = set(dirs)
        files = set(files)
        expected_dirs = set()
        expected_files = set()

        regression_test = []
        for item in currentnode["contents"]:
            if item["type"] == "directory":
                expected_dirs.add(item["name"])
            elif item["type"] == "file":
                expected_files.add(item["name"])
                if item.get("compare", False):
                    regression_test.append(item["name"])

        if expected_dirs != dirs:
            missing = expected_dirs - dirs
            unexpected = dirs - expected_dirs
            raise RuntimeError( \
                "Mismatched directories under '%s' found.\n" \
                "Missing : %s\n" \
                "Unexpected: %s" % (dirpath, missing, unexpected))

        if expected_files != files:
            missing = expected_files - files
            unexpected = files - expected_files
            raise RuntimeError( \
                "Mismatched files under '%s' found.\n" \
                "Missing : %s\n" \
                "Unexpected: %s" % (dirpath, missing, unexpected))

        if regression_test:
            from vtkmodules.vtkTestingRendering import vtkTesting

            t = vtkTesting()
            for image in regression_test:
                t.CleanArguments()
                t.AddArgument("-V")
                baseline_name = get_baseline(args.name, dirpath, image, args.root)
                t.AddArgument(os.path.join(args.baseline_dir, baseline_name))
                t.AddArgument("-T")
                t.AddArgument(args.temp_dir)
                if t.RegressionTest(os.path.join(dirpath, image), 0.05) != t.PASSED:
                    regressions_failure_count += 1
                    print("\n")
    if regressions_failure_count:
        raise RuntimeError("ERROR: %d regression tests failed!" % regressions_failure_count)


if __name__ == '__main__':
    main()



import paraview
from paraview import servermanager
from paraview.util import proxy as proxy_util

from paraview.modules.vtkRemotingCore import vtkPVSession

from paraview.simple import compatibility


# -----------------------------------------------------------------------------


def LoadPalette(paletteName):
    """Load a color palette to override the default foreground and background
    colors used by ParaView views. The current global palette's colors are set
    to the colors in the loaded palette.

    :param paletteName: Name of the palette to load from this list:
        'WarmGrayBackground', 'DarkGrayBackground', 'NeutralGrayBackground',
        'LightGrayBackground', 'WhiteBackground', 'BlackBackground',
        'GradientBackground'.
    :type paletteName: str"""
    name = compatibility.GetPaletteName(paletteName)

    pxm = servermanager.ProxyManager()
    palette = pxm.GetProxy("settings", "ColorPalette")
    prototype = pxm.GetPrototypeProxy("palettes", name)

    if palette is None or prototype is None:
        return

    palette.Copy(prototype)
    palette.UpdateVTKObjects()


# -----------------------------------------------------------------------------


def GetColorTransferFunction(arrayname, representation=None, separate=False, **params):
    """Get the color transfer function used to map a data array with the given name to
    colors.

    :param arrayname: Name of the array whose color transfer function is requested.
    :type arrayname: str
    :param representation: Used to modify the array name when using a separate color
        transfer function. Optional, defaults to the active proxy.
    :type representation: Representation proxy.
    :param separate: If `True`, used to recover the separate color transfer function
        even if it is not used currently by the representation. Optional, defaults to
        `False`.
    :type separate: bool
    :return: This may create a new color transfer function if none exists, or return an
        existing one if found.
    :rtype: Color transfer function proxy"""
    if representation:
        if separate or representation.UseSeparateColorMap:
            arrayname = "%s_%s_%s" % (
                "Separate",
                representation.SMProxy.GetGlobalIDAsString(),
                arrayname,
            )
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    lut = servermanager._getPyProxy(
        tfmgr.GetColorTransferFunction(arrayname, session.GetSessionProxyManager())
    )
    proxy_util.set(lut, **params)
    return lut


# -----------------------------------------------------------------------------


def GetBlockColorTransferFunction(
    selector, arrayname, representation=None, separate=False, **params
):
    """Get the color transfer function used to map a data array with the given name to
    colors in the blocks referred to by a selector expresson.

    :param selector: Selector expression used to select blocks from whom the color
        transfer should be retrieved.
    :type selector: str
    :param arrayname: Name of the array whose color transfer function is requested.
    :type arrayname: str
    :param representation: Used to modify the array name when using a separate color
        transfer function. Optional, defaults to the active proxy.
    :type representation: Representation proxy.
    :param separate: If `True`, used to recover the separate color transfer function
        even if it is not used currently by the representation. Optional, defaults to
        `False`.
    :type separate: bool
    :return: This may create a new color transfer function if none exists, or return an
        existing one if found.
    :rtype: Color transfer function proxy"""
    if representation:
        if representation.GetBlockUseSeparateColorMap(selector):
            arrayname = "%s_%s_%s_%s" % (
                "Separate_",
                representation.SMProxy.GetGlobalIDAsString(),
                selector,
                arrayname,
            )
        elif separate or representation.UseSeparateColorMap:
            arrayname = "%s_%s_%s" % (
                "Separate",
                representation.SMProxy.GetGlobalIDAsString(),
                arrayname,
            )
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    lut = servermanager._getPyProxy(
        tfmgr.GetColorTransferFunction(arrayname, session.GetSessionProxyManager())
    )
    proxy_util.set(lut, **params)
    return lut


# -----------------------------------------------------------------------------


def GetOpacityTransferFunction(
    arrayname, representation=None, separate=False, **params
):
    """Get the opacity transfer function used to map a data array with the given name to
    opacities in the blocks referred to by a selector expresson.

    :param arrayname: Name of the array whose opacity transfer function is requested.
    :type arrayname: str
    :param representation: Used to modify the array name when using a separate opacity
        transfer function. Optional, defaults to the active representation.
    :type representation: Representation proxy.
    :param separate: If `True`, used to recover the separate opacity transfer function
        even if it is not used currently by the representation. Optional, defaults to
        `False`.
    :type separate: bool
    :return: This may create a new opacity transfer function if none exists, or return an
        existing one if found.
    :rtype: Opacity transfer function proxy"""
    if representation:
        if separate or representation.UseSeparateColorMap:
            arrayname = "%s%s_%s" % (
                "Separate_",
                representation.SMProxy.GetGlobalIDAsString(),
                arrayname,
            )
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    otf = servermanager._getPyProxy(
        tfmgr.GetOpacityTransferFunction(arrayname, session.GetSessionProxyManager())
    )
    proxy_util.set(otf, **params)
    return otf


# -----------------------------------------------------------------------------


def GetTransferFunction2D(
    arrayname, array2name=None, representation=None, separate=False, **params
):
    """Get the 2D color transfer function used to map a data array with the given name to
    colors.

    :param arrayname: Name of the first array whose color transfer function is requested.
    :type arrayname: str
    :param array2name: Name of the second array whose color transfer function is requested.
    :type array2name: str
    :param representation: Used to modify the array name when using a separate color
        transfer function. Optional, defaults to the active representation.
    :type representation: Representation proxy.
    :param separate: If `True`, used to recover the separate color transfer function
        even if it is not used currently by the representation. Optional, defaults to
        `False`.
    :type separate: bool
    :return: This may create a new 2D color transfer function if none exists, or return an
        existing one if found.
    :rtype: 2D color transfer function proxy"""
    array2name = "" if array2name is None else ("_%s" % array2name)
    if representation:
        if separate or representation.UseSeparateColorMap:
            arrayname = "%s%s_%s%s" % (
                "Separate_",
                representation.SMProxy.GetGlobalIDAsString(),
                arrayname,
                array2name,
            )
    if not servermanager.ActiveConnection:
        raise RuntimeError("Missing active session")
    session = servermanager.ActiveConnection.Session
    tfmgr = servermanager.vtkSMTransferFunctionManager()
    tf2d = servermanager._getPyProxy(
        tfmgr.GetTransferFunction2D(arrayname, session.GetSessionProxyManager())
    )
    proxy_util.set(tf2d, **params)
    return tf2d


# -----------------------------------------------------------------------------


def ImportPresets(filename, location=vtkPVSession.CLIENT):
    """Import presets from a file. The file can be in the legacy color map XML
    format or in the new JSON format.

    :param filename: Path of the file containing the presets.
    :type filename: str
    :param location: Where the statefile should be save, e.g., pass `vtkPVSession.CLIENT`
        if the statefile is located on the client system (default value), pass in
        `vtkPVSession.SERVERS` if on the server. Optional, defaults to
        `vtkPVSession.CLIENT`.
    :type location: `vtkPVServer.ServerFlags` enum value
    :return: `True` on success, `False` otherwise."""
    presets = servermanager.vtkSMTransferFunctionPresets.GetInstance()
    return presets.ImportPresets(filename, location)


# -----------------------------------------------------------------------------
def ExportTransferFunction(
    colortransferfunction,
    opacitytransferfunction,
    tfname,
    filename,
    location=vtkPVSession.CLIENT,
):
    """Export transfer function to a file. The file will be saved in the new JSON format.

    :param colortransferfunction: The color transfer function to export.
    :type colortransferfunction: Color transfer function proxy.
    :param opacitytransferfunction: The opacity transfer functon to export, if provided.
        can be None.
    :type opacitytransferfunction: Opacity transfer function proxy.
    :param tfname: The name that will be given to the transfer function preset if imported
        back into ParaView.
    :type tfname: str
    :param filename: Path to file where exported transfer function should be saved.
    :type filename: str
    :param location: Where the statefile should be saved, e.g., pass `vtkPVSession.CLIENT`
        if the statefile is located on the client system (default value), pass in
        `vtkPVSession.SERVERS` if on the server. Optional, defaults to `vtkPVSession.CLIENT`.
    :type location: `vtkPVServer.ServerFlags` enum value
    :return: `True` on success, `False` otherwise."""
    return servermanager.vtkSMTransferFunctionProxy.ExportTransferFunction(
        colortransferfunction.SMProxy,
        opacitytransferfunction.SMProxy,
        tfname,
        filename,
        location,
    )


# -----------------------------------------------------------------------------


def CreateLookupTable(**params):
    """Create and return a lookup table.

    :param params: A variadic list of `key=value` pairs giving values of
        specific named properties of the lookup table.
    :return: Lookup table
    :rtype: Lookup table proxy"""
    lt = servermanager.rendering.PVLookupTable()
    controller = servermanager.ParaViewPipelineController()
    controller.InitializeProxy(lt)
    proxy_util.set(lt, **params)
    controller.RegisterColorTransferFunctionProxy(lt)
    return lt


# -----------------------------------------------------------------------------


def CreatePiecewiseFunction(**params):
    """Create and return a piecewise function.

    :param params: A variadic list of `key=value` pairs giving values of
        specific named properties of the opacity function.
    :return: Piecewise opacity function
    :rtype: Opacity functon proxy"""
    pfunc = servermanager.piecewise_functions.PiecewiseFunction()
    controller = servermanager.ParaViewPipelineController()
    controller.InitializeProxy(pfunc)
    proxy_util.set(pfunc, **params)
    controller.RegisterOpacityTransferFunction(pfunc)
    return pfunc


# -----------------------------------------------------------------------------


def AssignFieldToColorPreset(array_name, preset_name, range_override=None):
    """Assign a lookup table to an array by lookup table name.

    Example usage::

        AssignFieldToPreset("Temperature", "Cool to Warm")

    :param array_name: The array name for which we want to bind a given color preset.
    :type array_name: str
    :param preset_name: The name for the color preset.
    :type preset_name: str
    :param range_override: The range to use instead of the range of the array.
    :type range_override: 2-element tuple or list of floats
    :return: `True` on success, `False` otherwise.
    :rtype: bool"""

    lut = GetColorTransferFunction(array_name)
    if not lut.ApplyPreset(preset_name):
        return False

    if range_override:
        lut.RescaleTransferFunction(range_override)

    return True


# -----------------------------------------------------------------------------


def ListColorPresetNames():
    """Returns a list containing the currently available color transfer function
    preset names.

    :return: List of currently available transfer function preset names.
    :rtype: List of str"""
    presets = servermanager.vtkSMTransferFunctionPresets.GetInstance()
    return [
        presets.GetPresetName(index) for index in range(presets.GetNumberOfPresets())
    ]


# -----------------------------------------------------------------------------


def GenerateRGBPoints(preset_name=None, range_min=None, range_max=None):
    """Create and return a list of RGB points by using the provided arguments.
    The returned RGB points are intended to be used by transfer functions.

    This function is used by Python state files so that they do not have to
    write out every single RGB point for a transfer function if those RGB
    points can be easily re-created.

    :param preset_name: A preset name to apply, if any.
    :type preset_name: str or None
    :param range_min: The minimum value for rescaling the RGBPoints range
    :type range_min: float or None
    :param range_min: The maximum value for rescaling the RGBPoints range
    :type range_max: float or None
    :return: A list of RGBPoints which can be used in transfer functions
    :rtype: list[float]

    NOTE: this function might be used by old state files.
    As such, we must not break backward-compatibility with it.
    """
    # Create a default PVLookupTable
    lut = servermanager.rendering.PVLookupTable()

    # If the named LUT is not in the presets, see if it was one that was removed and
    # substitute it with the backwards compatibility helper
    presets = servermanager.vtkSMTransferFunctionPresets.GetInstance()
    if not presets.HasPreset(preset_name):
        (preset_name, reverse) = (
            paraview._backwardscompatibilityhelper.lookupTableUpdate(preset_name)
        )

    # If no preset name was found, return an empty list.
    if preset_name is not None and not presets.HasPreset(preset_name):
        return []

    # If a preset name was provided, apply the preset
    if preset_name is not None:
        lut.SMProxy.ApplyPreset(preset_name)

    any_rescale = any(x is not None for x in (range_min, range_max))
    if any_rescale:
        if range_min is None:
            range_min = lut.RGBPoints[0]
        if range_max is None:
            range_max = lut.RGBPoints[-4]

        lut.SMProxy.RescaleTransferFunction([range_min, range_max])

    return lut.RGBPoints



import logging


class VTKHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super(VTKHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        try:
            from vtkmodules.vtkCommonCore import vtkLogger
            msg = self.format(record)
            lvl = self.get_vtk_level(record.levelno)
            vtkLogger.Log( \
                lvl,
                record.filename,
                record.lineno,
                msg)

            from vtkmodules.vtkCommonCore import vtkOutputWindow as win
            outputWindow = win.GetInstance()
            if outputWindow:
                # do not duplicate on standard output
                prevMode = outputWindow.GetDisplayMode()
                outputWindow.SetDisplayModeToNever()

                if lvl == vtkLogger.VERBOSITY_ERROR:
                    lvlText = 'ERR: '
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayErrorText(fullMsg)
                elif lvl == vtkLogger.VERBOSITY_WARNING:
                    lvlText = 'WARN: '
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayWarningText(fullMsg)
                else:
                    fullMsg = f"{record.filename}:{record.lineno} {msg}\n"
                    outputWindow.DisplayText(fullMsg)

                outputWindow.SetDisplayMode(prevMode)

        except Exception:
            self.handleError(record)

    def get_vtk_level(self, level):
        from vtkmodules.vtkCommonCore import vtkLogger
        if level >= logging.ERROR:
            return vtkLogger.VERBOSITY_ERROR
        elif level >= logging.WARNING:
            return vtkLogger.VERBOSITY_WARNING
        elif level >= logging.INFO:
            return vtkLogger.VERBOSITY_INFO
        elif level >= logging.DEBUG:
            return vtkLogger.VERBOSITY_TRACE
        else:
            return vtkLogger.VERBOSITY_MAX


def get_level(vtklevel=None):
    """returns current log level used by vtkLogger"""
    from vtkmodules.vtkCommonCore import vtkLogger
    vtk_level = vtkLogger.GetCurrentVerbosityCutoff() if vtklevel is None else vtklevel
    if vtk_level == vtkLogger.VERBOSITY_ERROR:
        return logging.ERROR
    elif vtk_level == vtkLogger.VERBOSITY_WARNING:
        return logging.WARNING
    elif vtk_level == vtkLogger.VERBOSITY_INFO:
        return logging.INFO
    else:
        return logging.DEBUG



import paraview.web.venv  # noqa
import importlib
import os
import sys

TRAME_APP = None

if "--trame-app" in sys.argv:
    base_index = sys.argv.index("--trame-app")
    TRAME_APP = sys.argv[base_index + 1]
    # remove --trame-app TRAME_APP from args
    sys.argv.pop(base_index)
    sys.argv.pop(base_index)
else:
    TRAME_APP = os.environ.get("TRAME_APP", None)


def main():
    if TRAME_APP is None:
        print("Could not find environment variable TRAME_APP or --trame-app arg")
        return

    module = importlib.import_module(TRAME_APP)
    module.main()


if __name__ == "__main__":
    main()



r"""
    This module is a ParaViewWeb server application used only for automated testing
"""

import os

# import paraview modules.
from paraview.web import pv_wslink
from paraview.web import protocols as pv_protocols
from wslink import server

import argparse


# =============================================================================
# Create custom Pipeline Manager class to handle clients requests
# =============================================================================


class _TestServer(pv_wslink.PVServerProtocol):

    dataDir = os.getcwd()
    canFile = None
    authKey = "wslink-secret"
    dsHost = None
    dsPort = 11111
    rsHost = None
    rsPort = 11111
    rcPort = -1
    fileToLoad = None
    groupRegex = "[0-9]+\\.[0-9]+\\.|[0-9]+\\."
    excludeRegex = "^\\.|~$|^\\$"
    plugins = None
    filterFile = None
    colorPalette = None
    proxies = None
    allReaders = True
    saveDataDir = os.getcwd()
    viewportScale = 1.0
    viewportMaxWidth = 2560
    viewportMaxHeight = 1440

    def initialize(self):
        # Bring used components
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebStartupRemoteConnection(
                _TestServer.dsHost,
                _TestServer.dsPort,
                _TestServer.rsHost,
                _TestServer.rsPort,
                _TestServer.rcPort,
            )
        )
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebStartupPluginLoader(_TestServer.plugins)
        )
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebFileListing(
                _TestServer.dataDir,
                "Home",
                _TestServer.excludeRegex,
                _TestServer.groupRegex,
            )
        )
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebProxyManager(
                allowedProxiesFile=_TestServer.proxies,
                baseDir=_TestServer.dataDir,
                fileToLoad=_TestServer.fileToLoad,
                allowUnconfiguredReaders=_TestServer.allReaders,
            )
        )
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebColorManager(
                pathToColorMaps=_TestServer.colorPalette
            )
        )
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebMouseHandler())
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebViewPort()
        )  # _TestServer.viewportScale, _TestServer.viewportMaxWidth, _TestServer.viewportMaxHeight
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebViewPortImageDelivery())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebViewPortGeometryDelivery())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebTimeHandler())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebSelectionHandler())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebWidgetManager())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebKeyValuePairStore())
        self.registerVtkWebProtocol(
            pv_protocols.ParaViewWebSaveData(baseSavePath=_TestServer.saveDataDir)
        )

        # Update authentication key to use
        self.updateSecret(_TestServer.authKey)


# =============================================================================
# Main: Parse args and start server
# =============================================================================

if __name__ == "__main__":
    # Create argument parser
    parser = argparse.ArgumentParser(description="ParaView Web Test Server")

    # Add arguments
    server.add_arguments(parser)
    args = parser.parse_args()

    # Start server
    server.start_webserver(options=args, protocol=_TestServer)



r"""This module is used to export catalyst defined data products directly
within the ParaView GUI."""

from paraview.simple import *

haveCinemaC = True
try:
    import paraview.tpl.cinema_python.adaptors.paraview.pv_introspect as pvi
except:
    haveCinemaC = False
import os


class CinemaDHelper(object):
    """ A helper that we funnel file save commands through so that we can build up a CinemaD table for them. """

    def __init__(self, mcd, rd):
        self.__EnableCinemaDTable = mcd
        self.__RootDirectory = rd
        if rd and not rd.endswith("/"):
            self.__RootDirectory = rd + "/"
        self.Keys = set()
        self.Contents = []
        self.KeysWritten = None

    def __StripRootDir(self, filename):
        if self.__RootDirectory:
            return filename[filename.startswith(self.__RootDirectory) and len(self.__RootDirectory):]
        return filename

    def __MakeCinDFileNamesUnderRootDir(self, filename=""):
        """ filename wrangling for root directory placement """
        indexfilename = 'data.csv'
        datafilename = filename
        if self.__RootDirectory:
            indexfilename = self.__RootDirectory + "data.csv"
            # strip leading root directory from filename if present
            datafilename = self.__StripRootDir(filename)
        return indexfilename, datafilename

    def AppendToCinemaDTable(self, time, producer, filename):
        """ keep a record of every standard file written out so that we can list it later """
        if not self.__EnableCinemaDTable:
            return
        indexfilename, datafilename = self.__MakeCinDFileNamesUnderRootDir(filename)
        self.Keys.add("timestep")
        self.Keys.add("producer")
        self.Keys.add("FILE")
        self.Contents.append({'timestep': time, 'producer': producer, 'FILE': datafilename})

    def AppendCViewToCinemaDTable(self, time, producer, filelist):
        """ keep a record of every new file that cinema image writes along with the keys that produced them so that we can list them all later """
        if not self.__EnableCinemaDTable or not haveCinemaC:
            return
        self.Keys.add("timestep")
        self.Keys.add("producer")
        self.Keys.add("FILE")
        # unroll the contents into key lists and filenames
        for viewname in filelist:
            for entry in filelist[viewname]:
                keylist = entry[0]
                # avoid redundancy from name change
                if 'time' in keylist:
                    time = keylist['time']
                    del keylist['time']
                for k in keylist:
                    self.Keys.add(k)
                keylist['timestep'] = time
                keylist['producer'] = producer
                keylist['FILE'] = self.__StripRootDir(entry[1])
                self.Contents.append(keylist)

    def Finalize(self):
        """ Finish off the table and write it to a file """
        # this is necessary because we don't really know keys (from cinema image) until the end.
        if not self.__EnableCinemaDTable:
            return
        indexfilename, datafilename = self.__MakeCinDFileNamesUnderRootDir()
        if self.__RootDirectory and not os.path.exists(self.__RootDirectory):
            os.makedirs(self.__RootDirectory)
        f = open(indexfilename, "w")
        # write the header line
        f.write("timestep,")
        f.write("producer,")
        for k in self.Keys:
            if k != 'timestep' and k != 'producer' and k != 'FILE':
                f.write("%s," % k)
        f.write("FILE\n")
        # write all of the contents
        for l in self.Contents:
            f.write("%s," % l['timestep'])
            f.write("%s," % l['producer'])
            for k in self.Keys:
                if k != 'timestep' and k != 'producer' and k != 'FILE':
                    v = ''
                    if k in l:
                        v = l[k]
                    f.write("%s," % v)
            f.write("%s\n" % self.__StripRootDir(l['FILE']))
        f.close()

    def WriteNow(self):
        """ For Catalyst, we don't generally have a Final state, so we call this every CoProcess call and fixup the table if we have to. """
        if not self.__EnableCinemaDTable:
            return

        indexfilename, datafilename = self.__MakeCinDFileNamesUnderRootDir()
        if self.KeysWritten == self.Keys:
            # phew nothing new, we can just append a record
            f = open(indexfilename, "a+")
            for l in self.Contents:
                f.write("%s," % l['timestep'])
                f.write("%s," % l['producer'])
                for k in self.Keys:
                    if k != 'timestep' and k != 'producer' and k != 'FILE':
                        v = ''
                        if k in l:
                            v = l[k]
                        f.write("%s," % v)
                f.write("%s\n" % self.__StripRootDir(l['FILE']))
            f.close()
            self.Contents = []
            return
        # dang, whatever we wrote recently had a new variable
        # we may have to extend and rewrite the old output file
        readKeys = None
        readContents = []
        if os.path.exists(indexfilename):
            # yep we have to do it
            # parse the old file
            f = open(indexfilename, "r")
            for line in f:
                read = line[0:-1].split(",")  # -1 to skip trailing\n
                if readKeys is None:
                    readKeys = read
                else:
                    entry = {}
                    for idx in range(0, len(read)):
                        entry.update({readKeys[idx]: read[idx]})
                    readContents.append(entry)
        # combine contents
        if readKeys is not None:
            self.Keys = self.Keys.union(readKeys)
            readContents.extend(self.Contents)
            self.Contents = readContents
        # finally, write
        self.Finalize()
        self.KeysWritten = self.Keys.copy()
        self.Contents = []


class __CinemaACHelper(object):
    """ Another helper that connects up to cinema_python's export function. """

    def __init__(self, rd, vsel, tsel, asel):
        self.__RootDirectory = rd
        self.__ViewSelection = vsel
        self.__TrackSelection = tsel
        self.__ArraySelection = asel
        self.NewFiles = None

    if haveCinemaC:
        def ExportNow(self, time):
            r = pvi.export_scene(baseDirName=self.__RootDirectory,
                                 viewSelection=dict(self.__ViewSelection),
                                 trackSelection=dict(self.__TrackSelection),
                                 arraySelection=dict(self.__ArraySelection),
                                 forcetime=time)
            self.NewFiles = r
    else:
        def ExportNow(self, time):
            pass


def _fixup_and_makedir_if_needed(rootdir):
    if rootdir and not rootdir.endswith("/"):
        rootdir = rootdir + "/"
    if rootdir and not os.path.exists(rootdir):
        os.makedirs(rootdir)
    return rootdir


def ExportNow(image_root_directory,
              data_root_directory,
              file_name_padding,
              make_cinema_table,
              cinema_tracks,
              cinema_arrays,
              rendering_info):
    """The user facing entry point. Here we get a hold of ParaView's animation controls, step through the animation, and export the things we've been asked to be the caller."""

    CIND = CinemaDHelper(make_cinema_table, image_root_directory)
    image_root_directory = _fixup_and_makedir_if_needed(image_root_directory)
    data_root_directory = _fixup_and_makedir_if_needed(data_root_directory)

    # get a hold of the scene
    spm = servermanager.vtkSMProxyManager.GetProxyManager().GetActiveSessionProxyManager()
    ed = spm.GetExportDepot()
    s = GetAnimationScene()
    s.GoToFirst()
    et = s.EndTime
    tnow = s.AnimationTime

    numTimesteps = len(paraview.simple.GetAnimationScene().TimeKeeper.TimestepValues)
    timesteps = {}
    # We must perform the export loop at least once, even in the case where
    # the input data contains no timesteps
    if (numTimesteps == 0):
        timesteps = list(range(1))
    else:
        timesteps = list(range(numTimesteps))

    # export loop
    for tstep in timesteps:
        padded_tstep = str(tstep).rjust(file_name_padding, '0')
        helpers = []

        # loop through the configured writers and export at the requested times
        ed.InitNextWriterProxy()
        wp = ed.GetNextWriterProxy()
        writercnt = 0
        while wp:
            freq = wp.GetProperty("WriteFrequency").GetElement(0)
            if ((tstep % freq == 0) and
                    (not (wp.GetXMLName() == "Cinema image options"))):
                # this isn't pretty. I couldn't find a way to write
                # directly with the writer proxy. So what I do here
                # is find the input and filename and make a new writer
                # to use.
                proxyname = spm.GetProxyName("export_writers", wp)
                inputname = proxyname[0:proxyname.find("|")]
                inputproxy = FindSource(inputname)
                if wp.GetProperty("ChooseArraysToWrite").GetElement(0) == 1:
                    point_arrays = []
                    cell_arrays = []
                    arrays_property = wp.GetProperty("PointDataArrays")
                    for i in range(arrays_property.GetNumberOfElements()):
                        point_arrays.append(arrays_property.GetElement(i))
                    arrays_property = wp.GetProperty("CellDataArrays")
                    for i in range(arrays_property.GetNumberOfElements()):
                        cell_arrays.append(arrays_property.GetElement(i))
                    if not point_arrays:
                        point_arrays = [' ']
                    if not cell_arrays:
                        cell_arrays = [' ']
                    # create a temporary array culling filter
                    pass_arrays = PassArrays(Input=inputproxy, \
                                             PointDataArrays=point_arrays, CellDataArrays=cell_arrays)
                    inputproxy = pass_arrays
                    helpers.append(inputproxy)
                fname = wp.GetProperty("CatalystFilePattern").GetElement(0)
                if wp.GetXMLName() == "ExodusIIWriter":
                    fnamefilled = data_root_directory + fname + padded_tstep
                else:
                    fnamefilled = data_root_directory + fname.replace("%t", padded_tstep)

                kwargs = {}
                DataMode = wp.GetProperty("DataMode")
                if DataMode is not None:
                    kwargs["DataMode"] = wp.GetProperty("DataMode").GetElement(0)
                HeaderType = wp.GetProperty("HeaderType")
                if HeaderType is not None:
                    kwargs["HeaderType"] = wp.GetProperty("HeaderType").GetElement(0)
                EncodeAppendedData = wp.GetProperty("EncodeAppendedData")
                if EncodeAppendedData is not None:
                    kwargs["EncodeAppendedData"] = wp.GetProperty("EncodeAppendedData").GetElement(0)
                CompressorType = wp.GetProperty("CompressorType")
                if CompressorType is not None:
                    kwargs["CompressorType"] = wp.GetProperty("CompressorType").GetElement(0)
                CompressionLevel = wp.GetProperty("CompressionLevel")
                if CompressionLevel is not None:
                    kwargs["CompressionLevel"] = wp.GetProperty("CompressionLevel").GetElement(0)

                # finally after all of the finageling above, save the data
                SaveData(fnamefilled, inputproxy, **kwargs)
                # don't forget to tell cinema D about it
                CIND.AppendToCinemaDTable(tnow, "writer_%s" % writercnt, fnamefilled)
            wp = ed.GetNextWriterProxy()
            writercnt = writercnt + 1

        # loop through the configured screenshots and export at the requested times
        ed.InitNextScreenshotProxy()
        ssp = ed.GetNextScreenshotProxy()
        viewcnt = 0
        while ssp:
            if not ssp.HasAnnotation("enabled") or not (ssp.GetAnnotation("enabled") == '1'):
                ssp = ed.GetNextScreenshotProxy()
                continue
            freq = ssp.GetProperty("WriteFrequency").GetElement(0)
            if tstep % freq == 0:
                fname = ssp.GetProperty("CatalystFilePattern").GetElement(0)
                if fname.endswith("cdb"):
                    CINAC = __CinemaACHelper(image_root_directory,
                                             rendering_info,
                                             cinema_tracks,
                                             cinema_arrays)
                    # special treatment for cinema image data bases
                    CINAC.ExportNow(tnow)
                    # don't forget to tell cinema D about it
                    CIND.AppendCViewToCinemaDTable(tnow, "cview_%s" % viewcnt, CINAC.NewFiles)
                else:
                    fnamefilled = image_root_directory + fname.replace("%t", padded_tstep)
                    # save the screenshot
                    ssp.WriteImage(fnamefilled)
                    # don't forget to tell cinema D about it
                    CIND.AppendToCinemaDTable(tnow, "view_%s" % viewcnt, fnamefilled)
            ssp = ed.GetNextScreenshotProxy()
            viewcnt = viewcnt + 1
        tstep = tstep + 1
        s.GoToNext()
        tnow = s.AnimationTime

        # destroy array culling filters
        for x in helpers:
            Delete(x)

    # defer actual cinema D output until the end because we only know now what the full set of cinema D columns actually are
    CIND.Finalize()



# This script takes as input a Proxy XML file
# and return a cpp header file that Qt Linguist
# tools can itself transform into a Qt translation
# source file.
# For example, this XML file:
"""
<ServerManagerConfiguration>
    <ProxyGroup name="example">
        <SourceProxy label="Image Data To AMR"
                     name="ImageDataToAMR">
        </SourceProxy>
    </ProxyGroup>
</ServerManagerConfiguration>
"""
# would result in the following header:
"""
#include <QtGlobal>
#include <iostream>
#include <string>

static const char *messages[] = {
    //: Real source: example.xml:2 - example.xml
    QT_TRANSLATE_NOOP("ServerManagerXML", R"(example)"),

    //: Real source: example.xml:3 - example
    QT_TRANSLATE_NOOP("ServerManagerXML", R"(Image Data To AMR)")
};
"""

import sys
# Forbid the use of C modules so the parser overload can get the correct line numbers
sys.modules['_elementtree'] = None

import xml.etree.ElementTree as ET
import argparse
from collections.abc import Iterable
import os

def translationUnit(file: str, line: int, context: str, content: str) -> str:
    if not content:
        return ""
    content = ' '.join(content.replace('"', '\\"').split())
    # line is not used for now as it creates git merge conflict when modifying the translation and ParaView at the same time
    res = f"\t//: Real source: {file} - {context}\n"
    res += f"\tQT_TRANSLATE_NOOP(\"ServerManagerXML\", R\"({content})\"),\n\n"
    return res

def _insertSpace(string: str, index: int) -> str:
    return string[:index] + ' ' + string[index:]

# This function MUST have the same behavior as
# vtkSMObject::CreatePrettyLabel()
def createPrettyLabel(label: str) -> str:
    """
    Create a proper label from a name.
    """
    if label == "":
        return ""
    label = " " + label + "  "
    for i in range(0, len(label)):
        if label[i - 1].isspace() or label[i + 1].isspace():
            continue
        if label[i].isupper() and label[i + 1].islower():
            label = _insertSpace(label, i)
        if label[i].islower() and label[i + 1].isupper():
            label = _insertSpace(label, i + 1)
    return label[1:-2]

def recursiveStringCrawl(file: str, context: str, group: str, node) -> list:
    """
    Recursive function searching for node labels to translate.
    Return a list of its translatable attributes (and from its children)
    to translate as Qt header format.
    """

    res = []

    # Identify proxy group
    if node.tag == "ProxyGroup":
      group = node.attrib["name"]

    else:
      if "label" in node.attrib:
          res.append(translationUnit(file, node.line, context, node.attrib["label"]))
      elif "menu_label" in node.attrib:
          res.append(translationUnit(file, node.line, context, node.attrib["menu_label"]))
      elif "Documentation" in node.tag:
          if node.text:
              res.append(translationUnit(file, node.line, context, node.text))
          if "long_help" in node.attrib:
              res.append(translationUnit(file, node.line, context, node.attrib["long_help"]))
          if "short_help" in node.attrib:
              res.append(translationUnit(file, node.line, context, node.attrib["short_help"]))
      elif "Text" in node.tag:
          if node.text:
              res.append(translationUnit(file, node.line, context, node.text))
          if "title" in node.attrib:
              res.append(translationUnit(file, node.line, context, node.attrib["title"]))
      elif "Property" in node.tag and "name" in node.attrib:
          res.append(translationUnit(file, node.line, context, createPrettyLabel(node.attrib["name"])))
      elif node.tag.endswith("Proxy") and not group.startswith("internal_") and "name" in node.attrib:
          res.append(translationUnit(file, node.line, context, createPrettyLabel(node.attrib["name"])))
      elif "ShowInMenu" in node.tag and "category" in node.attrib:
          res.append(translationUnit(file, node.line, context, node.attrib["category"]))

      if node.tag.endswith("Proxy") and "name" in node.attrib:
          context = node.attrib["name"]

    for child in node:
        res += recursiveStringCrawl(file, context, group, child)

    return res

class LineParser(ET.XMLParser):
    """
    Overload default XML parser to give nodes a line attribute.
    """
    def _start(self, *args, **kwargs):
        el = super(self.__class__, self)._start(*args, **kwargs)
        el.line = self.parser.CurrentLineNumber
        return el

def removePrefix(string: str, prefix: str):
    if string.startswith(prefix):
        return string[len(prefix):]
    return string

def fileParsing(fileIn: str, sourceDir: str) -> str:
    """
    Starts recursiveStringCrawl at xml file root.
    Returns the string of all labels to translate as
    Qt header format.
    """
    xmlTree = ET.parse(fileIn, parser=LineParser())
    stringsList = recursiveStringCrawl(removePrefix(fileIn, sourceDir), removePrefix(fileIn, sourceDir), "root", xmlTree.getroot())
    res = "".join(stringsList)
    return res

def filesManager(filesIn: list, fileOut: str, sourceDir: str):
    """
    Create Qt header first lines then call
    fileParsing for each file and appends them
    to the result file.
    """
    if not isinstance(filesIn, Iterable):
        raise ValueError("Not an array")
    res = "#include <QtGlobal>\n#include <iostream>\n#include <string>\n\n"
    res += "static const char *messages[] = {\n"
    for file in filesIn:
        res += fileParsing(file, sourceDir)
    # Removing a comma and newline after last element of the "message" list.
    res = res[:-2]
    res += "\n};\n"
    with open(fileOut, 'w', encoding='utf-8') as f:
        f.write(res)

def __main__():
    """
    Asks for an ouput file and for one or more input files.
    """
    parser = argparse.ArgumentParser("Generates C++ headers with XML content to be translated")
    parser.add_argument('-o', '--out', help='Result header filename', nargs="?")
    parser.add_argument('-s', '--source', help='Client source directory path', nargs=1)
    parser.add_argument('inFiles', metavar='N', nargs="+")
    args = parser.parse_args()
    for i in args.inFiles:
        if not os.path.exists(i):
            raise ValueError(f"Invalid input file '{i}'")
    if args.out is not None and args.source is not None:
        filesManager(args.inFiles, args.out, args.source[0])
    else:
        for i in args.inFiles:
            filesManager([i], os.path.splitext(i)[0] + ".h")

if __name__ == "__main__":
    __main__()



from paraview import servermanager


# ==============================================================================
# Plugin Management
# ==============================================================================


def LoadXML(xmlstring, ns=None):
    """Given a server manager XML as a string, parse and process it.
    If you loaded the simple module with ``from paraview.simple import *``,
    make sure to pass ``globals()`` as the second arguments::

        LoadXML(xmlstring, globals())

    Otherwise, the new functions will not appear in the global namespace.

    :param xmlstring: XML string containing server manager definitions.
    :type xmlstring: str
    :param ns: Namespace in which new functions will be defined. Optional,
               defaults to `None`.
    :type ns: dict

    """
    from paraview.simple import _extend_simple

    if not ns:
        ns = globals()
    servermanager.LoadXML(xmlstring)
    _extend_simple(ns)


# -----------------------------------------------------------------------------


def LoadPlugin(filename, remote=True, ns=None):
    """Loads a ParaView plugin and updates this module with new constructors
    if any. The remote argument (default to ``True``) is to specify whether
    the plugin will be loaded on the client (``remote=False``) or on the
    server (``remote=True``).

    If you loaded the simple module with ``from paraview.simple import *``,
    make sure to pass ``globals()`` as an argument::

        LoadPlugin("myplugin", False, globals()) # to load on client
        LoadPlugin("myplugin", True, globals())  # to load on server
        LoadPlugin("myplugin", ns=globals())     # to load on server

    Otherwise, the new functions will not appear in the global namespace.

    :param filename: Path to the plugin to load.
    :type filename: str
    :param remote: If `True`, loads the plugin on the server unless the
        connection is not remote. If `False`, loads the plugin on the client.
        Optional, defaults to `True`.
    :type remote: bool
    :param ns: Namespace in which new functions will be loaded. Optional,
        defaults to `None`.
    :type ns: dict
    :return: None
    """
    return LoadPlugins(filename, remote=remote, ns=ns)


# -----------------------------------------------------------------------------


def LoadPlugins(*args, **kwargs):
    """Loads ParaView plugins and updates this module with new constructors
    if any. The remote keyword argument (default to ``True``) is to specify
    whether the plugin will be loaded on client (``remote=False``) or on server
    (``remote=True``). Proxy definition updates are deferred until all plugins
    have been read, which can be more computationally efficient when multiple
    plugins are loaded in sequence.

    If you loaded the simple module with ``from paraview.simple import *``,
    make sure to pass ``globals()`` as a keyword argument::

        LoadPlugins("myplugin1", "myplugin2", remote=False, ns=globals()) # to load on client
        LoadPlugins("myplugin", "myplugin2", remote=True, ns=globals())  # to load on server
        LoadPlugins("myplugin", "myplugin2", ns=globals())     # to load on server

    Otherwise, the new functions will not appear in the global namespace.

    Note, `remote=True` has no effect when the connection is not remote.
    """
    from paraview.simple import _extend_simple

    remote = True
    if "remote" in kwargs:
        remote = kwargs["remote"]

    ns = globals()
    if "ns" in kwargs and kwargs["ns"]:
        ns = kwargs["ns"]

    servermanager.vtkSMProxyManager.GetProxyManager().SetBlockProxyDefinitionUpdates(
        True
    )
    for arg in args:
        servermanager.LoadPlugin(arg, remote)
    servermanager.vtkSMProxyManager.GetProxyManager().SetBlockProxyDefinitionUpdates(
        False
    )
    servermanager.vtkSMProxyManager.GetProxyManager().UpdateProxyDefinitions()

    _extend_simple(ns)


# -----------------------------------------------------------------------------


def LoadDistributedPlugin(pluginname, remote=True, ns=None):
    """Loads a plugin that's distributed with the ParaView executable. This uses the
    information known about plugins distributed with ParaView to locate the
    shared library for the plugin to load.

    :param pluginname: Short name of the plugin (not a file path)
    :type pluginname: str
    :param remote: If `True`, loads the plugin on the server unless the
        connection is not remote. If `False`, loads the plugin on the client.
        Optional, defaults to `True`.
    :type remote: bool
    :param ns: Namespace in which new functions will be loaded. Optional,
        defaults to `None`.
    :type ns: dict
    :raises RuntimeError: If the plugin was not found.
    """
    if not servermanager.ActiveConnection:
        raise RuntimeError("Cannot load a plugin without a session.")

    conn = servermanager.ActiveConnection
    do_remote = remote and conn.IsRemote()

    plm = servermanager.vtkSMProxyManager.GetProxyManager().GetPluginManager()
    if do_remote:
        session = servermanager.ActiveConnection.Session
        info = plm.GetRemoteInformation(session)
    else:
        info = plm.GetLocalInformation()
    for cc in range(0, info.GetNumberOfPlugins()):
        if info.GetPluginName(cc) == pluginname:
            return LoadPlugin(info.GetPluginFileName(cc), remote=do_remote, ns=ns)
    raise RuntimeError("Plugin '%s' not found" % pluginname)


# ==============================================================================
# Custom Filters Management
# ==============================================================================
def LoadCustomFilters(filename, ns=None):
    """Loads a custom filter XML file and updates this module with new
    constructors if any. If you loaded the simple module with
    ``from paraview.simple import *``, make sure to pass ``globals()`` as an
    argument.

    :param filename: Path to XML file with custom filter definitions.
    :type filename: str
    :param ns: Namespace in which new functions will be loaded. Optional,
        defaults to `None`.
    :type ns: dict
    """
    from paraview.simple import _extend_simple

    servermanager.ProxyManager().SMProxyManager.LoadCustomProxyDefinitions(filename)
    if not ns:
        ns = globals()

    _extend_simple(ns)



# __init__.py file for cinema_viewers

# DO NOT REMOVE THIS FILE

#
# this file is included in the repository so that this repository can
# be included as a submodule of a python module
#
# it does nothing but report the version number

Version = "2.2"

def version():
    return Version





import os
import numpy

#
# channel class
#
class channel:
    """Composible Image Set channel class

    A channel is:
    - a name
    - a variable name
    - a variable min
    - a variable max
    - a variable type
    - a two dimensional array of values. 
    """
    def __init__(self):
        self.data = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def colormap(self):
        """ A colormap is a dictionary with at least one entry, 
            'points', which is a list of dicts each of which is expected 
            to have five elements:

            'r': r float value
            'g': g float value
            'b': b float value
            'x': x position float value
            'a': alpha float value

            If there is no entry "colorspace", then "rgb" is assumed.
        """
        return self._colormap

    @colormap.setter
    def colormap(self, value):
        self._colormap = value

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        self._active = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def var(self):
        return self._var

    @var.setter
    def var(self, value):
        self._var = value

    @property
    def varmin(self):
        return self._varmin

    @varmin.setter
    def varmin(self, value):
        self._varmin = value

    @property
    def varmax(self):
        return self._varmax

    @varmax.setter
    def varmax(self, value):
        self._varmax = value

    @property
    def vartype(self):
        return self._vartype

    @var.setter
    def vartype(self, value):
        self._vartype = value

    def load(self, url):
        self.url = url
        if os.path.isfile(url):
            zdata = numpy.load(url)
            self.data = zdata['data']



r"""Module for generating a Python state for ParaView.
This module uses paraview.smtrace to generate a trace for a selected set of
proxies by mimicking the creation of various pipeline components in sequence.
Typical usage of this module is as follows::

    from paraview import smstate
    state = smstate.get_state()
    print (state)

Note, this cannot be called when Python tracing is active.
"""

from paraview import servermanager as sm
from paraview import smtrace
from paraview import simple

RECORD_MODIFIED_PROPERTIES = sm.vtkSMTrace.RECORD_MODIFIED_PROPERTIES
RECORD_ALL_PROPERTIES = sm.vtkSMTrace.RECORD_ALL_PROPERTIES
RECORD_ACTIVE_MODIFIED_PROPERTIES = sm.vtkSMTrace.RECORD_ACTIVE_MODIFIED_PROPERTIES


class supported_proxies(object):
    """filter object used to hide proxies that are currently not supported by
    the state saving mechanism or those that are generally skipped in state.
    There is none currently but keeping the method for backward compatibility
    and future proofing."""

    def __call__(self, proxy):
        return proxy


class visible_representations(object):
    """filter object to skip hidden representations from being saved in state file"""

    def __call__(self, proxy):
        if not supported_proxies()(proxy): return False
        try:
            return proxy.Visibility
        except AttributeError:
            pass
        return True


def __toposort(input_set):
    """implementation of Tarjan topological sort to sort proxies using consumer
    dependencies as graph edges."""
    result = []
    marked_set = set()
    while marked_set != input_set:
        unmarked_node = (input_set - marked_set).pop()
        __toposort_visit(result, unmarked_node, input_set, marked_set)
    result.reverse()
    return result


def __toposort_visit(result, proxy, input_set, marked_set, t_marked_set=None):
    if t_marked_set is None:
        temporarily_marked_set = set()
    else:
        temporarily_marked_set = t_marked_set
    if proxy in temporarily_marked_set:
        raise RuntimeError("Cycle detected in pipeline! %r" % proxy)
    if not proxy in marked_set:
        temporarily_marked_set.add(proxy)
        consumers = set()
        get_consumers(proxy, lambda x: x in input_set, consumer_set=consumers, recursive=False)
        for x in consumers:
            __toposort_visit(result, x, input_set, marked_set, temporarily_marked_set)
        marked_set.add(proxy)
        temporarily_marked_set.discard(proxy)
        result.append(proxy)


def get_consumers(proxy, filter, consumer_set, recursive=True):
    """Returns the consumers for a proxy iteratively. If filter is non-None,
    filter is used to cull consumers."""
    for i in range(proxy.GetNumberOfConsumers()):
        consumer = proxy.GetConsumerProxy(i)
        consumer = consumer.GetTrueParentProxy() if consumer else None
        consumer = sm._getPyProxy(consumer)
        if not consumer or consumer.IsPrototype() or consumer in consumer_set:
            continue
        if filter(consumer):
            consumer_set.add(consumer)
            if recursive: get_consumers(consumer, filter, consumer_set)


def get_producers(proxy, filter, producer_set):
    """Returns the producers for a proxy iteratively. If filter is non-None,
    filter is used to cull producers."""
    for i in range(proxy.GetNumberOfProducers()):
        producer = proxy.GetProducerProxy(i)
        producer = producer.GetTrueParentProxy() if producer else None
        producer = sm._getPyProxy(producer)
        if not producer or producer.IsPrototype() or producer in producer_set:
            continue
        if filter(producer):
            producer_set.add(producer)
            get_producers(producer, filter, producer_set)
    # FIXME: LookupTable is missed :/, darn subproxies!
    try:
        if proxy.LookupTable and filter(proxy.LookupTable):
            producer_set.add(proxy.LookupTable)
            get_producers(proxy.LookupTable, filter, producer_set)
    except AttributeError:
        pass
    try:
        if proxy.ScalarOpacityFunction and filter(proxy.ScalarOpacityFunction):
            producer_set.add(proxy.ScalarOpacityFunction)
            get_producers(proxy.ScalarOpacityFunction, filter, producer_set)
    except AttributeError:
        pass


def get_state(options=None, source_set=[], filter=None, raw=False,
              preamble=None, postamble=None):
    """Returns the state string"""
    if options:
        options = sm._getPyProxy(options)
        propertiesToTraceOnCreate = options.PropertiesToTraceOnCreate.ConvertValue(options.PropertiesToTraceOnCreate.GetElement(0))
        skipHiddenRepresentations = options.SkipHiddenDisplayProperties
        skipRenderingComponents = options.SkipRenderingComponents
        skipActiveComponents = options.SkipActiveComponents
        skipLayoutComponents = options.SkipLayoutComponents
        skipAnimationComponents = options.SkipAnimationComponents
    else:
        propertiesToTraceOnCreate = RECORD_MODIFIED_PROPERTIES
        skipHiddenRepresentations = True
        skipRenderingComponents = False
        skipActiveComponents = False
        skipLayoutComponents = False
        skipAnimationComponents = False

    # essential to ensure any obsolete accessors don't linger - can cause havoc
    # when saving state following a Python trace session
    # (paraview/paraview#18994)
    import gc
    gc.collect()

    if sm.vtkSMTrace.GetActiveTracer():
        raise RuntimeError("Cannot generate Python state when tracing is active.")

    if filter is None:
        filter = visible_representations() if skipHiddenRepresentations else supported_proxies()

    # build a set of proxies of interest
    if source_set:
        start_set = source_set
    else:
        # if nothing is specified, we save all views, sources, selections and animation scene
        start_set = [x for x in simple.GetSources().values()]
        start_set += [x for x in simple.GetSelections().values()]
        start_set += simple.GetViews()
        start_set += [simple.GetAnimationScene()]

    start_set = [x for x in start_set if filter(x)]

    # now, locate dependencies for the start_set, pruning irrelevant branches
    consumers = set(start_set)
    for proxy in start_set:
        get_consumers(proxy, filter, consumers)

    producers = set()
    for proxy in consumers:
        get_producers(proxy, filter, producers)

    # proxies_of_interest is set of all proxies that we should trace.
    proxies_of_interest = producers.union(consumers)

    tracer = smtrace.ScopedTracer(preamble="")
    with tracer:
        # this ensures that lookup tables/scalar bars etc. are fully traced.
        tracer.config.SetFullyTraceSupplementalProxies(True)
        tracer.config.SetSkipRenderingComponents(skipRenderingComponents)
        tracer.config.SetPropertiesToTraceOnCreate(propertiesToTraceOnCreate)

        trace = smtrace.TraceOutput()
        if preamble is None:
            trace.append("# state file generated using %s" % simple.GetParaViewSourceVersion())
            trace.append("import paraview")
            trace.append("paraview.compatibility.major = %d" % sm.vtkSMProxyManager.GetVersionMajor())
            trace.append("paraview.compatibility.minor = %d" % sm.vtkSMProxyManager.GetVersionMinor())
        elif preamble:
            trace.append(preamble)
        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # --------------------------------------------------------------------------
        # We trace the views and layouts, if any.
        if skipRenderingComponents:
            views = []
        else:
            views = [x for x in proxies_of_interest if smtrace.Trace.get_registered_name(x, "views")]

        if views:
            # sort views by their names, so the state has some structure to it.
            views = sorted(views, key=lambda x: \
                smtrace.Trace.get_registered_name(x, "views"))
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup views used in the visualization",
                "# ----------------------------------------------------------------"])
            for view in views:
                # FIXME: save view camera positions and size.
                traceitem = smtrace.RegisterViewProxy(view)
                traceitem.finalize()
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

            if not skipActiveComponents:
                trace.append_separated(["SetActiveView(None)"])

        # from views,  build the list of layouts of interest.
        layouts = set()
        for aview in views:
            l = simple.GetLayout(aview)
            if l:
                layouts.add(simple.GetLayout(aview))

        # trace create of layouts
        if layouts and not skipLayoutComponents:
            layouts = sorted(layouts, key=lambda x: \
                smtrace.Trace.get_registered_name(x, "layouts"))
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup view layouts",
                "# ----------------------------------------------------------------"])
            for layout in layouts:
                traceitem = smtrace.RegisterLayoutProxy(layout)
                traceitem.finalize(filter=lambda x: x in views)
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        if views and not skipActiveComponents:
            # restore the active view after the layouts have been created.
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# restore active view",
                "SetActiveView(%s)" % smtrace.Trace.get_accessor(simple.GetActiveView()),
                "# ----------------------------------------------------------------"])

        # --------------------------------------------------------------------------
        # Next, trace selection sources.
        sorted_proxies_of_interest = __toposort(proxies_of_interest)
        sorted_selection_sources = [x for x in sorted_proxies_of_interest \
                             if smtrace.Trace.get_registered_name(x, "selection_sources")]

        if sorted_selection_sources:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup the selection sources",
                "# ----------------------------------------------------------------"])
            for selection in sorted_selection_sources:
                traceitem = smtrace.RegisterSelectionProxy(selection)
                traceitem.finalize("selection_sources")
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # --------------------------------------------------------------------------
        # Next, trace selections.
        sorted_selections = [x for x in sorted_proxies_of_interest \
                             if smtrace.Trace.get_registered_name(x, "selections")]
        selected_source_id = None
        if sorted_selections and len(sorted_selections) > 1:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup the selections",
                "# ----------------------------------------------------------------"])
            for selection in sorted_selections:
                traceitem = smtrace.RegisterSelectionProxy(selection)
                traceitem.finalize("selections")
                if selection.GetXMLName() == "AppendSelections":
                    selected_source_id = selection.GetSelectionId()
                    selected_source_port = selection.GetSelectionPort()
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # --------------------------------------------------------------------------
        # Next, trace data processing pipelines.
        sorted_sources = [x for x in sorted_proxies_of_interest \
                          if smtrace.Trace.get_registered_name(x, "sources")]
        if sorted_sources:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup the data processing pipelines",
                "# ----------------------------------------------------------------"])
            for source in sorted_sources:
                traceitem = smtrace.RegisterPipelineProxy(source, saving_state=True)
                traceitem.finalize()
                if source.GetGlobalID() == selected_source_id:
                    selected_source_varname = traceitem.varname
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # --------------------------------------------------------------------------
        # Link selections to selected source
        if sorted_selections and len(sorted_selections) > 1:
            trace.append_separated([ \
                f"appendSelections.SetSelectionId({selected_source_varname}.GetGlobalID())",
                f"appendSelections.SetSelectionPort({selected_source_port})"])

        # --------------------------------------------------------------------------
        # Can't decide if the representations should be saved with the pipeline
        # objects or afterwards, opting for afterwards for now since the topological
        # sort doesn't guarantee that the representations will follow their sources
        # anyways.
        sorted_representations = [x for x in sorted_proxies_of_interest \
                                  if smtrace.Trace.get_registered_name(x, "representations")]
        scalarbar_representations = [x for x in sorted_proxies_of_interest \
                                     if smtrace.Trace.get_registered_name(x, "scalar_bars")]
        # print ("sorted_representations", sorted_representations)
        # print ("scalarbar_representations", scalarbar_representations)
        if not skipRenderingComponents and (sorted_representations or scalarbar_representations):
            for view in views:
                view_representations = [x for x in view.Representations if x in sorted_representations]
                view_scalarbars = [x for x in view.Representations if x in scalarbar_representations]
                if view_representations or view_scalarbars:
                    trace.append_separated([ \
                        "# ----------------------------------------------------------------",
                        "# setup the visualization in view '%s'" % smtrace.Trace.get_accessor(view),
                        "# ----------------------------------------------------------------"])
                for rep in view_representations:
                    try:
                        producer = rep.Input
                        port = rep.Input.Port
                        traceitem = smtrace.Show(producer, port, view, rep,
                                                 comment="show data from %s" % smtrace.Trace.get_accessor(producer))
                        traceitem.finalize()
                        del traceitem
                        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

                        if rep.UseSeparateColorMap:
                            trace.append_separated([ \
                                "# set separate color map",
                                "%s.UseSeparateColorMap = True" % ( \
                                    smtrace.Trace.get_accessor(rep))])

                        if rep.BlockUseSeparateColorMaps:
                            selectorsAndUseSeparateColorMap = rep.BlockUseSeparateColorMaps.GetData()
                            if len(selectorsAndUseSeparateColorMap) % 2 == 0 and len(selectorsAndUseSeparateColorMap) > 0:
                                trace.append_separated([ \
                                    "# set separate color map for blocks",
                                    "%s.BlockUseSeparateColorMaps = %s" % ( \
                                        smtrace.Trace.get_accessor(rep), ", ".join(selectorsAndUseSeparateColorMap))])
                    except AttributeError:
                        pass
                # save the scalar bar properties themselves.
                if view_scalarbars:
                    trace.append_separated("# setup the color legend parameters for each legend in this view")
                    for rep in view_scalarbars:
                        smtrace.Trace.get_accessor(rep)
                        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))
                        trace.append_separated([ \
                            "# set color bar visibility", "%s.Visibility = %s" % ( \
                                smtrace.Trace.get_accessor(rep), rep.Visibility)])

                for rep in view_representations:
                    try:
                        producer = rep.Input
                        port = rep.Input.Port

                        if rep.IsScalarBarVisible(view):
                            # FIXME: this will save this multiple times, right now,
                            # if two representations use the same LUT.
                            # If someone wants to fix it, also fix it below for blocks.
                            trace.append_separated([ \
                                "# show color legend",
                                "%s.SetScalarBarVisibility(%s, True)" % ( \
                                    smtrace.Trace.get_accessor(rep),
                                    smtrace.Trace.get_accessor(view))])

                        if rep.BlockColorArrayNames:
                            selectors = rep.BlockColorArrayNames.GetData()[::3]
                            isScalarVisibles = rep.IsBlocksScalarBarVisible(view, selectors)
                            if all(isScalarVisible == 1 for isScalarVisible in isScalarVisibles):
                                trace.append_separated([ \
                                    "# set color map for blocks",
                                    "%s.SetBlocksScalarBarVisibility(%s, %s, True)" % ( \
                                        smtrace.Trace.get_accessor(rep),
                                        smtrace.Trace.get_accessor(view),
                                        str(selectors))])
                            else:
                                for i in range(len(selectors)):
                                    if isScalarVisibles[i]:
                                        trace.append_separated([ \
                                            "# set color map for block %s" % selectors[i],
                                            "%s.SetBlockScalarBarVisibility(%s, %s, True)" % ( \
                                                smtrace.Trace.get_accessor(rep),
                                                smtrace.Trace.get_accessor(view),
                                                selectors[i])])

                        if not rep.Visibility:
                            traceitem = smtrace.Hide(producer, port, view)
                            traceitem.finalize()
                            del traceitem
                            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

                    except AttributeError:
                        pass

        # --------------------------------------------------------------------------
        # Now, trace the transfer functions (color maps and opacity maps) used.
        ctfs = set([x for x in proxies_of_interest if smtrace.Trace.get_registered_name(x, "lookup_tables")])
        if not skipRenderingComponents and ctfs:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup color maps and opacity maps used in the visualization",
                "# note: the Get..() functions create a new object, if needed",
                "# ----------------------------------------------------------------"])
            for ctf in ctfs:
                smtrace.Trace.get_accessor(ctf)
                if ctf.ScalarOpacityFunction in proxies_of_interest:
                    smtrace.Trace.get_accessor(ctf.ScalarOpacityFunction)
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # --------------------------------------------------------------------------
        # Now, trace the animation scenes
        # There should be only one but keep the code generic
        anims = set([x for x in proxies_of_interest \
                     if smtrace.Trace.get_registered_name(x, "animation")])
        if not skipRenderingComponents and anims and not skipAnimationComponents:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup animation scene, tracks and keyframes",
                "# note: the Get..() functions create a new object, if needed",
                "# ----------------------------------------------------------------"])
            for anim in anims:
                traceitem = smtrace.TraceAnimationProxy(anim)
                traceitem.finalize()
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # Trace extractors.
        exgens = set([x for x in proxies_of_interest \
                      if smtrace.Trace.get_registered_name(x, "extractors")])
        if exgens:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# setup extractors",
                "# ----------------------------------------------------------------"])
            for exgen in exgens:
                # FIXME: this currently doesn't handle multiple output ports
                # correctly.
                traceitem = smtrace.CreateExtractor( \
                    xmlname=exgen.Writer.GetXMLName(),
                    producer=exgen.Producer,
                    extractor=exgen,
                    registrationName=smtrace.Trace.get_registered_name(exgen, "extractors"))
                traceitem.finalize()
                del traceitem
            trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

        # restore the active source since the order in which the pipeline is created
        # in the state file can end up changing the active source to be different
        # than what it was when the state is being saved.
        if not skipActiveComponents:
            trace.append_separated([ \
                "# ----------------------------------------------------------------",
                "# restore active source",
                "SetActiveSource(%s)" % smtrace.Trace.get_accessor(simple.GetActiveSource()),
                "# ----------------------------------------------------------------"])

        if postamble is None:
            trace.append_separated(smtrace._get_standard_postamble_comment())
        elif postamble:
            trace.append_separated(postamble)

        return str(trace) if not raw else trace.raw_data()

    return ""


if __name__ == "__main__":
    print("Running test")
    simple.Mandelbrot()
    simple.Show()
    simple.Hide()
    simple.Shrink().ShrinkFactor = 0.4
    simple.UpdatePipeline()
    simple.Clip().ClipType.Normal[1] = 1

    rep = simple.Show()
    view = simple.Render()
    view.ViewSize = [500, 500]
    rep.SetScalarBarVisibility(view, True)
    simple.Render()
    #    rep.SetScalarBarVisibility(view, False)

    print("====================================================================")
    print(get_state())



"""
Module that looks at a ParaView pipeline and automatically creates a cinema
store that ranges over all of the variables that we know how to control and
later show.
"""
from __future__ import absolute_import

from ...database import store, file_store
from ...adaptors import explorers
from ...adaptors.paraview import pv_explorers
from ...adaptors.paraview import progress

try:
    from itertools import imap
except ImportError:
    # python 3
    imap = map

def py23iteritems(d):
    myit = None
    try:
        myit = d.iteritems()
    except AttributeError:
        myit = d.items()
    return myit

import math
import os.path
import paraview
import copy


def record_visibility():
    """
    at start of run, record the current paraview state so we can return to it
    """
    proxies = []

    view_info = {}
    view_proxy = paraview.simple.GetActiveView()

    view_info['proxy'] = "__view_info"
    view_info[
        'orientation_axis_visibility'] = view_proxy.OrientationAxesVisibility

    camera = view_proxy.GetActiveCamera()
    view_info['position'] = camera.GetPosition()
    view_info['view_up'] = camera.GetViewUp()
    view_info['focal_point'] = camera.GetFocalPoint()
    proxies.append(view_info)

    source_proxies = paraview.servermanager.ProxyManager().GetProxiesInGroup(
        "sources")

    for key in source_proxies:
        listElt = {}
        proxy = source_proxies[key]
        # TODO: Find a way to record widget visibility so it can be restored
        paraview.simple.Hide3DWidgets(proxy=proxy)
        listElt['proxy'] = proxy
        listElt['visibility'] = None
        listElt['scalar_bar_visibility'] = False
        listElt['color_array_name'] = None
        listElt['color_array_association'] = None
        rep = paraview.servermanager.GetRepresentation(proxy, view_proxy)
        if rep is not None:
            listElt['visibility'] = rep.Visibility
            listElt['scalar_bar_visibility'] = rep.IsScalarBarVisible(
                view_proxy)
            listElt['color_array_name'] = rep.ColorArrayName.GetArrayName()
            listElt['color_array_association'] = \
                rep.ColorArrayName.GetAssociation()
        proxies.append(listElt)
    return proxies


def max_bounds():
    """ returns conservative min and max (over x y and z) bounds """
    source_proxies = paraview.servermanager.ProxyManager().GetProxiesInGroup(
        "sources")

    minb = 0
    maxb = -1
    for key in source_proxies:
        proxy = source_proxies[key]
        bounds = proxy.GetDataInformation().GetBounds()
        if bounds[0] < minb:
            minb = bounds[0]
        if bounds[2] < minb:
            minb = bounds[2]
        if bounds[4] < minb:
            minb = bounds[4]
        if bounds[1] > maxb:
            maxb = bounds[1]
        if bounds[3] > maxb:
            maxb = bounds[3]
        if bounds[5] > maxb:
            maxb = bounds[5]
    db = maxb-minb
    minb = minb-db
    maxb = maxb+db
    return minb, maxb


def restore_visibility(proxies):
    """at end of run, return to a previously recorded paraview state"""
    view_proxy = paraview.simple.GetActiveView()

    for listElt in proxies:
        if listElt['proxy'] == "__view_info":
            view_proxy.OrientationAxesVisibility = listElt[
                'orientation_axis_visibility']

            camera = view_proxy.GetActiveCamera()
            camera.SetPosition(listElt['position'])
            camera.SetViewUp(listElt['view_up'])
            camera.SetFocalPoint(listElt['focal_point'])
        else:
            proxy = listElt['proxy']
            vis = listElt['visibility']
            if vis is not None:
                rep = paraview.servermanager.GetRepresentation(
                    proxy, view_proxy)
                if rep is not None:
                    rep.Visibility = listElt['visibility']
                    if listElt['color_array_association']:
                        rep.SetScalarColoring(
                            listElt['color_array_name'],
                            paraview.servermanager.GetAssociationFromString(
                                listElt['color_array_association']))
                    if listElt['scalar_bar_visibility']:
                        rep.SetScalarBarVisibility(
                            view_proxy, listElt['scalar_bar_visibility'])


def inspect(skip_invisible=True):
    """
    Produces a representation of the pipeline that is easier to work with.
    Thanks Scott Wittenburg and the pv mailing list for this gem
    """
    source_proxies = paraview.servermanager.ProxyManager().GetProxiesInGroup(
        "sources")
    view_proxy = paraview.simple.GetActiveView()
    proxies = []
    proxybyId = {}
    for key in source_proxies:
        listElt = {}
        listElt['name'] = key[0]
        listElt['id'] = key[1]
        proxy = source_proxies[key]

        # skip the invisible
        rep = paraview.servermanager.GetRepresentation(proxy, view_proxy)
        if skip_invisible:
            if rep is None:
                # for example, writers in catalyst pipeline
                continue

        listElt['visibility'] = 0 if rep is None else rep.Visibility

        parentId = '0'
        try:
            if hasattr(proxy, 'Input'):
                parentId = proxy.Input.GetGlobalIDAsString()
        except AttributeError:
            parentId = '0'
        listElt['parent'] = parentId
        proxies.append(listElt)
        proxybyId[key[1]] = listElt

    if skip_invisible:
        # reparent upward over invisible parents
        for l in proxies:
            pid = l['parent']
            if pid not in proxybyId:
                pid = '0'
            while pid != '0' and proxybyId[pid]['visibility'] == 0:
                pid = proxybyId[pid]['parent']
            l['parent'] = pid

        # remove invisible proxies themselves
        pxies = []
        for l in proxies:
            if l['visibility'] != 0:
                pxies.append(l)
    else:
        pxies = proxies

    return pxies


def get_pipeline():
    """sanitizes the pipeline graph"""
    proxies = inspect(skip_invisible=False)
    for proxy in proxies:
        source = paraview.simple.FindSource(proxy['name'])
        numberOfProducers = source.GetNumberOfProducers()
        if proxy['parent'] == '0' and numberOfProducers > 0:
            # this proxy is the result of a merge
            parents = []
            for i in xrange(numberOfProducers):
                parents.append(
                    source.GetProducerProxy(i).GetGlobalIDAsString())
            proxy['parents'] = parents
        else:
            proxy['parents'] = [proxy['parent']]
        del proxy['parent']
    for proxy in proxies:
        proxy['children'] = [p['id'] for p in proxies
                             if proxy['id'] in p['parents']]
    return proxies


def float_limiter(x):
    """a shame, but needed to make sure python, javascript and
    (directory/file)name agree. TODO: This can go away now that
    we use name=index instead of name=value filenames."""
    if isinstance(x, (float)):
        return '%.6e' % x  # arbitrarily chose 6 significant digits
    else:
        return x


# Keeps a link between a filter and its explorer-track.
# Populated in addFilterValue()
# and queried in explore()
explorerDir = {}


def add_filter_value(name, cs, userDefinedValues):
    """creates controls for the filters that we know how to manipulate"""
    source = paraview.simple.FindSource(name)

    # Determine the filter type
    valueType = None
    if isinstance(source, paraview.simple.servermanager.filters.Clip):
        valueType = 'OffsetValues'
        explorerDir[name] = pv_explorers.Clip(name, source)
    elif isinstance(source, paraview.simple.servermanager.filters.Slice):
        valueType = 'SliceOffsetValues'
        explorerDir[name] = pv_explorers.Slice(name, source)
    elif isinstance(source, paraview.simple.servermanager.filters.Contour):
        valueType = 'Isosurfaces'
        explorerDir[name] = pv_explorers.Contour(name, source)

    if valueType is not None:
        # grab values from ui
        values = []
        if (source in userDefinedValues):
            if (valueType in userDefinedValues[source]):
                values = userDefinedValues[source][valueType]

        if len(values) == 0:
            # nothing asked for just leave as is
            return False

        # add sublayer and create the appropriate track
        cs.add_control(
            name,
            store.make_parameter(name, values, typechoice='hidden'))
        return True
    else:
        return False


def filter_has_parameters(name):
    """see if this proxy is one we know how to make controls for"""
    source = paraview.simple.FindSource(name)
    return any(imap(lambda filter: isinstance(source, filter),
                    [paraview.simple.servermanager.filters.Clip,
                     paraview.simple.servermanager.filters.Slice,
                     paraview.simple.servermanager.filters.Contour]))


def add_control_and_colors(name, cs, userDefined, arrayRanges, disableValues):
    """add parameters that change the settings and color of a filter"""
    source = paraview.simple.FindSource(name)
    # make up list of color options
    fields = {'depth': 'depth'}
    if not disableValues:
        fields['luminance'] = 'luminance'
    defaultName = None
    view_proxy = paraview.simple.GetActiveView()
    rep = paraview.servermanager.GetRepresentation(source, view_proxy)

    # select value arrays
    if rep.Representation != 'Outline':
        defaultName = add_customized_array_selection(
            name, source, fields, userDefined, arrayRanges, disableValues)

    if defaultName is None:
        fields['white'] = 'rgb'
        defaultName = 'white'

    cparam = store.make_field(
        "color"+name, fields, default=defaultName, valueRanges=arrayRanges)
    cs.add_field("color"+name, cparam, 'vis', [name])


def add_customized_array_selection(
        sourceName, source, fields, userDefined, arrayRanges, disableValues):
    def isArrayNotSelected(aName, arrays):
        return (aName not in arrays)

    defaultName = None
    if (source not in userDefined):
        return defaultName

    if ("arraySelection" not in userDefined[source]):
        return defaultName

    arrayNames = userDefined[source]["arraySelection"]

    cda = source.GetCellDataInformation()
    pda = source.GetPointDataInformation()
    for datainfo in [cda, pda]:
        for a in range(0, datainfo.GetNumberOfArrays()):
            arr = datainfo.GetArray(a)
            arrName = arr.GetName()

            if isArrayNotSelected(arrName, arrayNames):
                continue
            for i in range(0, arr.GetNumberOfComponents()):
                fName = arrName+"_"+str(i)
                if disableValues:
                    fields[fName] = 'lut'
                else:
                    fields[fName] = 'value'
                extend_range(arrayRanges, fName, list(arr.GetRange(i)))
                if defaultName is None:
                    defaultName = fName

    return defaultName


def range_epsilon(minmax):
    """
    ensure that min and max have some separation to assist rendering
    """
    if minmax[0] == minmax[1]:
        epsilon = minmax[0]*1E-6
        if epsilon == 0.0:
            epsilon = 1E-6
        minV = minmax[0] - epsilon
        maxV = minmax[1] + epsilon
        return [minV, maxV]
    return minmax


def extend_range(arrayRanges, name, minmax):
    """
    This updates the data ranges in the data base meta file.
    Throughout a time varying data export ranges will vary.
    Here we accumulate them as we go so that by the end we get
    the min and max values over for each array component over
    all time.

    This version happens in catalyst, where we recreate the
    database file every timestep.
    """
    adjustedMinMax = range_epsilon(minmax)
    if name in arrayRanges:
        temporalMinMax = list(arrayRanges[name])
        if adjustedMinMax[0] < temporalMinMax[0]:
            temporalMinMax[0] = adjustedMinMax[0]
        if adjustedMinMax[1] > temporalMinMax[1]:
            temporalMinMax[1] = adjustedMinMax[1]
        arrayRanges[name] = temporalMinMax
    else:
        arrayRanges[name] = adjustedMinMax


def update_all_ranges(cs, arrayRanges):
    """
    This updates the data ranges in the data base meta file.
    Throughout a time varying data export ranges will vary.
    Here we accumulate them as we go so that by the end we get
    the min and max values over for each array component over
    all time.

    This version happens in paraview export for time varying data
    where we recreate the database file once, and afterward only
    output new rasters.
    """
    plist = cs.parameter_list
    for pname, param in plist.items():
        if 'valueRanges' in param:
            # now we know it is a color type parameter
            filtername = pname[5:]  # get the filter name for this parameter
            proxy = paraview.simple.FindSource(filtername)  # get the filter
            if proxy is not None:
                for name, vrange in param['valueRanges'].items():
                    # iterate over all the colors choiced to get each array
                    # component
                    lrange = list(vrange)
                    compidx = name.rfind('_')  # get array name and component
                    aname = name[:compidx]
                    component = int(name[compidx+1:])

                    cai = proxy.GetCellDataInformation().GetArray(aname)
                    pai = proxy.GetPointDataInformation().GetArray(aname)
                    for arrinfo in [cai, pai]:
                        if arrinfo:
                            updated = False
                            drange = range_epsilon(arrinfo.GetRange(component))
                            if drange[0] < vrange[0]:
                                updated = True
                                lrange[0] = drange[0]
                            if drange[1] > vrange[1]:
                                updated = True
                                lrange[1] = drange[1]
                            if updated:
                                param['valueRanges'][name] = lrange


def optimize_traversal_order(params):
    """
    An optimization that reduces generation time by better amortizing
    setup costs. With this rendering state changes (cheap) are late
    and in particular camera (~free) is the innermost loop.

    It also makes generation easier to watch. With camera innermost
    it iunderstand and prevents rapidly cycling
    between bright colors.
    """
    try:
        params.remove("vis")
        params.append("vis")
    except (ValueError):
        pass

    constant = copy.deepcopy(params)
    for x in constant:
        if x.find("color") == 0:
            params.remove(x)
            params.append(x)

    try:
        params.remove("phi")
        params.append("phi")
        params.remove("theta")
        params.append("theta")
    except (ValueError):
        pass

    try:
        params.remove("pose")
        params.append("pose")
    except (ValueError):
        pass

    return params


def make_cinema_store(proxies,
                      ocsfname,
                      view,
                      forcetime=False,
                      userDefined={},
                      specLevel="A",
                      camType='phi-theta',
                      arrayRanges={},
                      extension=".png",
                      disableValues=False):

    """
    Takes in the pipeline, structured as a tree, and makes a cinema store
    definition containing all the parameters we will vary.
    """

    phis = userDefined.get('phi', [0, 180, 360])
    thetas = userDefined.get('theta', [0, 90, 180])
    rolls = userDefined.get('roll', [0, 45, 90, 135, 180, 225, 270, 315])

    if camType == 'static' or camType == 'phi-theta':
        rolls = [0]

    tvalues = []
    cs = file_store.FileStore(ocsfname)

    try:
        cs.load()
        tprop = cs.get_parameter('time')
        tvalues = tprop['values']

        # start with clean slate, other than time
        cs = file_store.FileStore(ocsfname)
    except (IOError, KeyError):
        pass

    eye_values = cs.metadata.get('camera_eye', [])
    at_values = cs.metadata.get('camera_at', [])
    up_values = cs.metadata.get('camera_up', [])
    nearfar_values = cs.metadata.get('camera_nearfar', [])
    viewangle_values = cs.metadata.get('camera_angle', [])

    cs.add_metadata({'store_type': 'FS'})
    if specLevel == "A":
        cs.add_metadata({'type': 'parametric-image-stack'})
        cs.add_metadata({'version': '0.0'})
    if specLevel == "B":
        cs.add_metadata({'type': 'composite-image-stack'})
        cs.add_metadata({'version': '0.2'})
    pipeline = get_pipeline()
    cs.add_metadata({'pipeline': pipeline})
    cs.add_metadata({'camera_model': camType})
    cs.add_metadata({'camera_eye': eye_values})
    cs.add_metadata({'camera_at': at_values})
    cs.add_metadata({'camera_up': up_values})
    cs.add_metadata({'camera_nearfar': nearfar_values})
    cs.add_metadata({'camera_angle': viewangle_values})

    vis = [proxy['name'] for proxy in proxies]
    if specLevel != "A":
        cs.add_layer("vis", store.make_parameter('vis', vis))

    pnames = []
    for proxy in proxies:
        proxy_name = proxy['name']
        ret = add_filter_value(proxy_name, cs, userDefined)
        if specLevel == "A" and ret:
            pnames.append(proxy_name)
        dependency_set = set([proxy['id']])
        repeat = True
        while repeat:
            repeat = False
            deps = set(proxy['id'] for proxy in proxies
                       if proxy['parent'] in dependency_set)
            if deps - dependency_set:
                dependency_set = dependency_set.union(deps)
                repeat = True
        dependency_list = [proxy['name'] for proxy in proxies
                           if proxy['id'] in dependency_set]
        if specLevel != "A":
            cs.assign_parameter_dependence(proxy_name, 'vis', dependency_list)
            add_control_and_colors(proxy_name, cs, userDefined, arrayRanges,
                                   disableValues)
            cs.assign_parameter_dependence(
                "color"+proxy_name, 'vis', [proxy_name])

    fnp = ""
    if forcetime is not False:
        # time specified, use it, being careful to append if already a list
        tvalues.append(forcetime)
        tprop = store.make_parameter('time', tvalues)
        cs.add_parameter('time', tprop)
        fnp = "{time}"
    else:
        # time not specified, try and make them automatically
        times = paraview.simple.GetAnimationScene().TimeKeeper.TimestepValues
        if times:
            prettytimes = [float_limiter(t) for t in times]
            cs.add_parameter(
                "time", store.make_parameter('time', prettytimes))
            fnp = "{time}"

    if camType == "phi-theta":
        bestp = phis[len(phis)//2]
        bestt = thetas[len(thetas)//2]
        cs.add_parameter(
            "phi",
            store.make_parameter('phi', phis,
                                 default=bestp))
        cs.add_parameter(
            "theta",
            store.make_parameter('theta', thetas,
                                 default=bestt))
        if fnp == "":
            fnp = "{phi}/{theta}"
        else:
            fnp = fnp + "/{phi}/{theta}"

    elif camType != "static":
        # for AER and YPR, make up a set of view matrices corresponding
        # to the requested number of samples in each axis
        def MatrixMul(mtx_a, mtx_b):
            tpos_b = zip(*mtx_b)
            rtn = [[sum(ea*eb for ea, eb in zip(a, b)) for b in tpos_b]
                   for a in mtx_a]
            return rtn

        poses = []  # holds phi, theta and roll angle tuples
        matrices = []  # holds corresponding transform matrices

        v = rolls[0]
        rolls = []
        if v < 2:
            rolls.append(0)
        else:
            j = -180
            while j < 180:
                rolls.append(j)
                j = j+360/v

        v = thetas[0]
        thetas = []
        if v < 2:
            thetas.append(0)
        else:
            j = -90
            while j <= 90:
                thetas.append(j)
                j = j+180/v

        for r in rolls:
            for t in thetas:
                v = phis[0]
                if v < 2:
                    poses.append((0, t, r))
                else:
                    # sample longitude less frequently toward the pole
                    increment_Scale = math.cos(math.pi*t/180.0)
                    if increment_Scale == 0:
                        increment_Scale = 1
                    # increment_Scale = 1  # for easy comparison
                    p = -180
                    while p < 180:
                        poses.append((p, t, r))
                        p = p+360/(v*increment_Scale)

        # default is one closest to 0,0,0
        dist = math.sqrt((poses[0][0]*poses[0][0]) +
                         (poses[0][1]*poses[0][1]) +
                         (poses[0][2]*poses[0][2]))
        default_mat = 0
        for i in poses:
            p, t, r = i
            cP = math.cos(-math.pi*(p/180.0))  # phi is right to left
            sP = math.sin(-math.pi*(p/180.0))
            cT = math.cos(-math.pi*(t/180.0))  # theta is up down
            sT = math.sin(-math.pi*(t/180.0))
            cR = math.cos(-math.pi*(r/180.0))  # roll is around gaze direction
            sR = math.sin(-math.pi*(r/180.0))
            rY = [[cP, 0, sP], [0, 1, 0], [-sP, 0, cP]]  # x,z interchange
            rX = [[1, 0, 0], [0, cT, -sT], [0, sT, cT]]  # y,z interchange
            rZ = [[cR, -sR, 0], [sR, cR, 0], [0, 0, 1]]  # x,y interchange
            m1 = [[1, 0, 0],  [0, 1, 0],  [0, 0, 1]]
            m2 = MatrixMul(m1, rY)
            m3 = MatrixMul(m2, rX)
            m4 = MatrixMul(m3, rZ)
            matrices.append(m4)
            newdist = math.sqrt(p*p+t*t+r*r)
            if newdist < dist:
                default_mat = m4
                dist = newdist

        cs.add_parameter("pose",
                         store.make_parameter('pose', matrices,
                                              default=default_mat))
        fnp = fnp+"{pose}"

    if specLevel == "A":
        for pname in pnames:
            if fnp == "":
                fnp = "{"+pname+"}"
            else:
                fnp = fnp+"/{"+pname+"}"

    if fnp == "":
        fnp = "image"

    cs.filename_pattern = fnp+extension
    return cs


def track_source(proxy, eye, at, up):
    """
    an animation mode that follows a specific object
    input camera position is in eye, at, up
    returns same, moved to follow the input proxy
    """
    # code duplicated from vtkPVCameraCueManipulator
    if proxy is None:
        return eye, at, up

    info = proxy.GetDataInformation()
    bounds = info.GetBounds()

    center = [(bounds[0] + bounds[1]) * 0.5,
              (bounds[2] + bounds[3]) * 0.5,
              (bounds[4] + bounds[5]) * 0.5]

    ret_eye = [center[0] + (eye[0] - at[0]),
               center[1] + (eye[1] - at[1]),
               center[2] + (eye[2] - at[2])]

    ret_at = [center[0], center[1], center[2]]
    return ret_eye, ret_at, up


def project_to_at(eye, fp, cr):
    """project center of rotation onto focal point to keep gaze direction the same
    while allowing both translate and zoom in and out to work"""
    d_fp = [fp[0]-eye[0], fp[1]-eye[1], fp[2]-eye[2]]
    d_cr = [cr[0]-eye[0], cr[1]-eye[1], cr[2]-eye[2]]
    num = (d_fp[0]*d_cr[0] + d_fp[1]*d_cr[1] + d_fp[2]*d_cr[2])
    den = (d_fp[0]*d_fp[0] + d_fp[1]*d_fp[1] + d_fp[2]*d_fp[2])
    if den == 0:
        return cr
    rat = num/den
    p_fp = [rat*d_fp[0], rat*d_fp[1], rat*d_fp[2]]
    at = [p_fp[0]+eye[0], p_fp[1]+eye[1], p_fp[2]+eye[2]]
    return at


def explore(cs, proxies, iSave=True, currentTime=None, userDefined={},
            specLevel="A",
            camType='phi-theta',
            tracking={},
            floatValues=True,
            arrayRanges={},
            disableValues=False,
            progressObject=None):
    """
    Runs a pipeline through all the changes we know how to make and saves off
    images into the store for each one.
    """

    view_proxy = paraview.simple.GetActiveView()
    dist = paraview.simple.GetActiveCamera().GetDistance()

    # associate control points with parameters of the data store
    params = list(cs.parameter_list.keys())
    tracks = []
    if camType == "phi-theta":
        up = [math.fabs(x) for x in view_proxy.CameraViewUp]
        uppest = 0
        if up[1] > up[uppest]:
            uppest = 1
        if up[2] > up[uppest]:
            uppest = 2
        cinup = [0, 0, 0]
        cinup[uppest] = 1

        eye = [x for x in view_proxy.CameraPosition]
        _fp = [x for x in view_proxy.CameraFocalPoint]
        _cr = [x for x in view_proxy.CenterOfRotation]
        at = project_to_at(eye, _fp, _cr)

        cam = pv_explorers.Camera(
            at, cinup, dist, view_proxy)
        tracks.append(cam)
    elif camType != 'static':
        cam = pv_explorers.PoseCamera(view_proxy, camType, cs)
        tracks.append(cam)

    cols = []

    ctime_float = None
    if currentTime:
        ctime_float = float(currentTime['time'])

    # hide all annotations
    view_proxy.OrientationAxesVisibility = 0

    for x in proxies:
        name = x['name']
        for y in params:

            if (y in explorerDir) and (name == y):
                # print ("name in ExplorerDir: ", y, ", ", explorerDir[y])
                tracks.append(explorerDir[y])

            if name in y:
                # visibility of the layer
                sp = paraview.simple.FindSource(name)
                if specLevel != "A":
                    rep = paraview.servermanager.GetRepresentation(
                        sp, view_proxy)

                    # hide all annotations
                    if rep.LookupTable:
                        rep.SetScalarBarVisibility(view_proxy, False)
                    tc1 = pv_explorers.SourceProxyInLayer(name, rep, sp)
                    lt = explorers.Layer('vis', [tc1])
                    tracks.append(lt)

                    # fields for the layer
                    cC = pv_explorers.ColorList()
                    cC.AddDepth('depth')
                    if not disableValues:
                        cC.AddLuminance('luminance')

                sp.UpdatePipeline(ctime_float)

                if specLevel != "A":
                    numVals = 0
                    if rep.Representation != 'Outline':
                        numVals = explore_customized_array_selection(
                            name, sp, cC, userDefined, disableValues)

                    if numVals == 0:
                        cC.AddSolidColor('white', [1, 1, 1])
                    col = pv_explorers.Color("color"+name, cC, rep)
                    tracks.append(col)
                    cols.append(col)

    params = optimize_traversal_order(params)

    explo = pv_explorers.ImageExplorer(cs, params,
                                       tracks,
                                       view_proxy,
                                       iSave)
    explo.enableFloatValues(floatValues)

    for c in cols:
        c.imageExplorer = explo

    eye_values = cs.metadata['camera_eye']
    at_values = cs.metadata['camera_at']
    up_values = cs.metadata['camera_up']
    nearfar_values = cs.metadata['camera_nearfar']
    viewangle_values = cs.metadata['camera_angle']

    eye = [x for x in view_proxy.CameraPosition]
    _fp = [x for x in view_proxy.CameraFocalPoint]
    _cr = [x for x in view_proxy.CenterOfRotation]
    at = project_to_at(eye, _fp, _cr)
    up = [x for x in view_proxy.CameraViewUp]
    times = paraview.simple.GetAnimationScene().TimeKeeper.TimestepValues
    if currentTime:
        times = False

    cam = paraview.simple.GetActiveCamera()

    # if tracking is turned on, find out how to move
    tracked_source = None
    if 'object' in tracking:
        # for now, just emulate animation's best mode with a mode that follows
        # an object
        objname = tracking['object']
        tracked_source = paraview.simple.FindSource(objname)
        if tracked_source is None:
            name_upper = objname[0].upper() + objname[1:]
            tracked_source = paraview.simple.FindSource(name_upper)

    if not times:
        eye, at, up = track_source(tracked_source, eye, at, up)
        try:
            tprop = cs.get_parameter('time')
            tvalues = tprop['values']
            for i in range(len(tvalues)):
                eye_values.append([x for x in eye])
                at_values.append([x for x in at])
                up_values.append([x for x in up])
                nearfar_values.append([x for x in cam.GetClippingRange()])
                viewangle_values.append(cam.GetViewAngle())
        except (KeyError):
            eye_values.append([x for x in eye])
            at_values.append([x for x in at])
            up_values.append([x for x in up])
            nearfar_values.append([x for x in cam.GetClippingRange()])
            viewangle_values.append(cam.GetViewAngle())
        cs.add_metadata({'camera_eye': eye_values})
        cs.add_metadata({'camera_at': at_values})
        cs.add_metadata({'camera_up': up_values})
        cs.add_metadata({'camera_nearfar': nearfar_values})
        cs.add_metadata({'camera_angle': viewangle_values})
        explo.explore(currentTime, progressObject)
    else:
        for t in times:
            view_proxy.ViewTime = t
            paraview.simple.Render(view_proxy)
            minbds, maxbds = max_bounds()
            view_proxy.MaxClipBounds = [
                minbds, maxbds, minbds, maxbds, minbds, maxbds]
            eye, at, up = track_source(tracked_source, eye, at, up)
            eye_values.append([x for x in eye])
            at_values.append([x for x in at])
            up_values.append([x for x in up])
            nearfar_values.append([x for x in cam.GetClippingRange()])
            viewangle_values.append(cam.GetViewAngle())

            cs.add_metadata({'camera_eye': eye_values})
            cs.add_metadata({'camera_at': at_values})
            cs.add_metadata({'camera_up': up_values})
            cs.add_metadata({'camera_nearfar': nearfar_values})
            cs.add_metadata({'camera_angle': viewangle_values})

            update_all_ranges(cs, arrayRanges)
            explo.explore({'time': float_limiter(t)}, progressObject)

    return cs.get_new_files()


def explore_customized_array_selection(
        sourceName, source, colorList, userDefined, disableValues):
    def isArrayNotSelected(aName, arrays):
        return (aName not in arrays)

    numVals = 0
    if (source not in userDefined):
        return numVals

    if ("arraySelection" not in userDefined[source]):
        return numVals

    arrayNames = userDefined[source]["arraySelection"]

    cda = source.GetCellDataInformation()
    pda = source.GetPointDataInformation()

    for datainfo, datatype in zip([cda, pda], ['CELLS', 'POINTS']):
        for a in range(0, datainfo.GetNumberOfArrays()):
            arr = datainfo.GetArray(a)
            arrName = arr.GetName()

            if isArrayNotSelected(arrName, arrayNames):
                continue
            for i in range(0, arr.GetNumberOfComponents()):
                numVals += 1
                if disableValues:
                    nforcomp = str(i)
                    if arr.GetNumberOfComponents() < 4:
                        nforcomp = 'X'
                        if i == 1:
                            nforcomp = 'Y'
                        if i == 2:
                            nforcomp = 'Z'
                    colorList.AddLUT(datatype, arrName+"_"+str(i), nforcomp)
                else:
                    isfield = (datatype == 'CELLS')
                    colorList.AddValueRender(
                        arrName + "_" + str(i),
                        isfield,
                        arrName,
                        i, arr.GetRange(i))

    return numVals


def export_scene(baseDirName, viewSelection, trackSelection, arraySelection, forcetime=False):
    '''
    This explores a set of user-defined views and tracks. export_scene is
    called from vtkCinemaExport.  The expected order of parameters is as
    follows:

    - viewSelection (following the format defined in
      Wrapping/Python/paraview/cpstate.py):

    Directory  of the form {'ViewName' : [parameters], ...}, with parameters
    defined in the order: Image filename, freq, fittoscreen, magnification,
    width, height, cinema).

    - trackSelection:

    Directory of the form {'FilterName' : [v1, v2, v3], ...}

    - arraySelection:

    Directory of the form {'FilterName' : ['arrayName1', 'arrayName2', ...],
                           ... }

    Note:  baseDirName is used as the parent directory of the database
    generated for each view in viewSelection. 'Image filename' is used as the
    database directory name.
    '''
    # save initial state
    initialView = paraview.simple.GetActiveView()
    pvstate = record_visibility()

    # a conservative global bounds for consistent z scaling
    minbds, maxbds = max_bounds()

    atLeastOneViewExported = False
    cinema_dirs = []
    new_files = {}
    for viewName, viewParams in py23iteritems(viewSelection):

        extension = os.path.splitext(viewParams[0])[1]

        # check if this view was selected to export as spec b
        cinemaParams = viewParams[6]
        if len(cinemaParams) == 0:
            print ("Skipping view: Not selected to export to cinema")
            continue

        camType = "none"
        if "camera" in cinemaParams and cinemaParams["camera"] != "none":
            camType = cinemaParams["camera"]
        if camType == "none":
            print ("Skipping view: Not selected to export to cinema.")
            continue

        specLevel = "A"
        if "composite" in cinemaParams and cinemaParams["composite"] is True:
            specLevel = "B"

        # get the view and save the initial status
        view = paraview.simple.FindView(viewName)
        paraview.simple.SetActiveView(view)
        view.ViewSize = [viewParams[4], viewParams[5]]

        view.MaxClipBounds = [minbds, maxbds, minbds, maxbds, minbds, maxbds]
        view.LockBounds = 1

        fitToScreen = viewParams[2]
        if fitToScreen != 0:
            if view.IsA("vtkSMContextViewProxy") is True:
                view.ResetDisplay()
            elif view.IsA("vtkSMRenderViewProxy") is False:
                print (' do not know what to do with a ', view.GetClassName())

        userDefValues = prepare_selection(trackSelection, arraySelection)
        if "theta" in cinemaParams:
            userDefValues["theta"] = cinemaParams["theta"]

        if "phi" in cinemaParams:
            userDefValues["phi"] = cinemaParams["phi"]

        if "roll" in cinemaParams:
            userDefValues["roll"] = cinemaParams["roll"]

        tracking_def = {}
        if "tracking" in cinemaParams:
            tracking_def = cinemaParams['tracking']

        # generate file path
        viewFileName = viewParams[0]
        viewDirName = viewFileName[0:viewFileName.rfind("_")]  # strip _num.ext
        filePath = os.path.join(baseDirName, viewDirName, "info.json")
        cinema_dirs.append(viewDirName)

        p = inspect()

        arrayRanges = {}
        disableValues = cinemaParams.get('noValues', False)

        cs = make_cinema_store(p, filePath, view, forcetime=forcetime,
                               userDefined=userDefValues,
                               specLevel=specLevel,
                               camType=camType,
                               arrayRanges=arrayRanges,
                               extension=extension,
                               disableValues=disableValues)
        enableFloatVal = cinemaParams.get('floatValues', False)

        pm = paraview.servermanager.vtkProcessModule.GetProcessModule()
        pid = pm.GetPartitionId()

        progObj = progress.ProgressObject()
        progObj.StartEvent()
        currentTime = None
        if forcetime is not False:
            currentTime = {'time':forcetime}
        explore(cs, p, iSave=(pid == 0),
                currentTime=currentTime,
                userDefined=userDefValues,
                specLevel=specLevel,
                camType=camType,
                tracking=tracking_def, floatValues=enableFloatVal,
                arrayRanges=arrayRanges, disableValues=disableValues,
                progressObject=progObj)
        progObj.EndEvent()

        new_files[viewName] = cs.get_new_files()

        view.LockBounds = 0

        if pid == 0:
            cs.save()
        atLeastOneViewExported = True

    if not atLeastOneViewExported:
        print ("No view was selected to export to cinema.")
        return

    make_workspace_file(baseDirName, cinema_dirs)

    # restore initial state
    paraview.simple.SetActiveView(initialView)
    restore_visibility(pvstate)
    return new_files


def prepare_selection(trackSelection, arraySelection):
    '''
    The rest of pv_introspect expects to receive user-defined values in the
    structure:

    {proxy_reference: {'ControlName': [value_1, value_2, ..., value_n],
                       'arraySelection': ['ArrayName_1', ..., 'ArrayName_n']}}

    This structure is necessary for catalyst to correctly reference the created
    proxies.  Although this is not necessary in the menu->export case (proxies
    could be accessed by name directly), we comply for compatibility.
    '''
    userDef = {}
    for name, values in py23iteritems(trackSelection):
        source = paraview.simple.FindSource(name)
        if (source is None):
            # Following the smtrace.py convention pqCinemaTrackSelection passes
            # lower-case-initial names, here the method tries to resolve the
            # upper-case-initial version of the name. Caveat: breaks if the
            # user re-names two items such that the only difference is the
            # first letter's capitalization (which would be confusing anyway).
            name_upper = name[0].upper() + name[1:]
            source = paraview.simple.FindSource(name_upper)

        if (source):
            options = userDef[source] if (source in userDef) else {}

            # Assumption: only a single 'ControlName' is supported per filter.
            # (the control name will need to be included in the ui query when
            # giving support to more parameters).
            controlName = ""
            if ("servermanager.Slice" in source.__class__().__str__() and
                "Plane object" in source.__getattribute__(
                    "SliceType").__str__()):
                controlName = "SliceOffsetValues"
            elif ("servermanager.Clip" in source.__class__().__str__() and
                  "Plane object" in
                  source.__getattribute__("ClipType").__str__()):
                controlName = "OffsetValues"
            elif ("servermanager.Contour" in source.__class__().__str__()):
                controlName = "Isosurfaces"

            if len(controlName) > 0:
                options[controlName] = values
                userDef[source] = options

    for name, arrayNames in py23iteritems(arraySelection):
        source = paraview.simple.FindSource(name)
        if (source is None):
            # Following the smtrace.py convention pqCinemaTrackSelection passes
            # lower-case-initial names, here the method tries to resolve the
            # upper-case-initial version of the name. Caveat: breaks if the
            # user re-names two items such that the only difference is the
            # first letter's capitalization (which would be confusing anyway).
            name_upper = name[0].upper() + name[1:]
            source = paraview.simple.FindSource(name_upper)

        if (source):
            options = userDef[source] if (source in userDef) else {}
            options["arraySelection"] = arrayNames
            userDef[source] = options

    return userDef


def make_workspace_file(basedirname, cinema_dirs):
    """
    This writes out the top level json file that says that there are
    child cinema stores inside. The viewer sees this and opens up
    in the children in separate panels.
    """
    pm = paraview.servermanager.vtkProcessModule.GetProcessModule()
    pid = pm.GetPartitionId()
    if len(cinema_dirs) > 1 and pid == 0:
        workspace = open(basedirname + 'info.json', 'w')
        workspace.write('{\n')
        workspace.write('    "metadata": {\n')
        workspace.write('        "type": "workbench"\n')
        workspace.write('    },\n')
        workspace.write('    "runs": [\n')
        for i in range(0, len(cinema_dirs)):
            workspace.write('        {\n')
            workspace.write('        "title": "%s",\n' % cinema_dirs[i])
            workspace.write('        "description": "%s",\n' % cinema_dirs[i])
            workspace.write('        "path": "%s"\n' % cinema_dirs[i])
            if i+1 < len(cinema_dirs):
                workspace.write('        },\n')
            else:
                workspace.write('        }\n')
                workspace.write('    ]\n')
                workspace.write('}\n')
                workspace.close()



try: paraview.simple
except: from paraview.simple import *

cp_writers = []

def RequestDataDescription(datadescription):
    "Callback to populate the request for current timestep"
    timestep = datadescription.GetTimeStep()


    input_name = 'input'
    if (timestep % 2 == 0) :
        datadescription.GetInputDescriptionByName(input_name).AllFieldsOn()
        datadescription.GetInputDescriptionByName(input_name).GenerateMeshOn()
    else:
        datadescription.GetInputDescriptionByName(input_name).AllFieldsOff()
        datadescription.GetInputDescriptionByName(input_name).GenerateMeshOff()


def DoCoProcessing(datadescription):
    "Callback to do co-processing for current timestep"
    global cp_writers
    cp_writers = []
    timestep = datadescription.GetTimeStep()

    PythonAdaptorDriver_vti = CreateProducer( datadescription, "input" )

    Slice1 = Slice( guiName="Slice1", SliceOffsetValues=[0.0], SliceType="Plane" )
    Slice1.SliceType.Offset = 0.0
    Slice1.SliceType.Origin = [0.25, 0.5, 0.5]
    Slice1.SliceType.Normal = [0.0, 1.0, 0.0]

    ParallelPolyDataWriter1 = CreateWriter( XMLPPolyDataWriter, "filename_{time}.pvtp", 2 )



    for writer in cp_writers:
        if timestep % writer.cpFrequency == 0:
            writer.FileName = writer.cpFileName.replace("%t", str(timestep))
            writer.UpdatePipeline()

    if timestep % 1 == 0:
        renderviews = servermanager.GetRenderViews()
        imagefilename = ""
        for view in range(len(renderviews)):
            fname = imagefilename.replace("%v", str(view))
            fname = fname.replace("%t", str(timestep))
            SaveScreenshot(fname, renderviews[view])

    # explicitly delete the proxies -- we do it this way to avoid problems with prototypes
    tobedeleted = GetProxiesToDelete()
    while len(tobedeleted) > 0:
        Delete(tobedeleted[0])
        tobedeleted = GetProxiesToDelete()

def GetProxiesToDelete():
    iter = servermanager.vtkSMProxyIterator()
    iter.Begin()
    tobedeleted = []
    while not iter.IsAtEnd():
      if iter.GetGroup().find("prototypes") != -1:
         iter.Next()
         continue
      proxy = servermanager._getPyProxy(iter.GetProxy())
      proxygroup = iter.GetGroup()
      iter.Next()
      if proxygroup != 'timekeeper' and proxy != None and proxygroup.find("pq_helper_proxies") == -1 :
          tobedeleted.append(proxy)

    return tobedeleted

def CreateProducer(datadescription, gridname):
  "Creates a producer proxy for the grid"
  if not datadescription.GetInputDescriptionByName(gridname):
    raise RuntimeError("Simulation input name '%s' does not exist" % gridname)
  grid = datadescription.GetInputDescriptionByName(gridname).GetGrid()
  producer = TrivialProducer()
  producer.GetClientSideObject().SetOutput(grid)
  producer.UpdatePipeline()
  return producer


def CreateWriter(proxy_ctor, filename, freq):
    global cp_writers
    writer = proxy_ctor()
    writer.FileName = filename
    writer.add_attribute("cpFrequency", freq)
    writer.add_attribute("cpFileName", filename)
    cp_writers.append(writer)
    return writer
